# coding: utf-8

"""
    CampaignManagementService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union, Set
from openapi_client.models.campaign.ad_recommendation_image_suggestion import AdRecommendationImageSuggestion
from openapi_client.models.campaign.ad_recommendation_video_suggestion import AdRecommendationVideoSuggestion
from openapi_client.models.campaign.responsive_ad import ResponsiveAd
from typing_extensions import Self

class CreateResponsiveAdRecommendationResponse(BaseModel):
    """
    CreateResponsiveAdRecommendationResponse
    """ # noqa: E501
    responsive_ad: Optional[ResponsiveAd] = Field(default=None, alias="ResponsiveAd")
    image_suggestions: Optional[List[Optional[AdRecommendationImageSuggestion]]] = Field(default=None, alias="ImageSuggestions")
    video_suggestions: Optional[List[Optional[AdRecommendationVideoSuggestion]]] = Field(default=None, alias="VideoSuggestions")
    prompt_brand_warning: Optional[StrictStr] = Field(default=None, alias="PromptBrandWarning")
    __properties: ClassVar[List[str]] = ["ResponsiveAd", "ImageSuggestions", "VideoSuggestions", "PromptBrandWarning"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )
	

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of responsive_ad
        if self.responsive_ad:
            _dict['ResponsiveAd'] = self.responsive_ad.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in image_suggestions (list)
        _items = []
        if self.image_suggestions:
            for _item_image_suggestions in self.image_suggestions:
                if _item_image_suggestions:
                    _items.append(_item_image_suggestions.to_dict())
            _dict['ImageSuggestions'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in video_suggestions (list)
        _items = []
        if self.video_suggestions:
            for _item_video_suggestions in self.video_suggestions:
                if _item_video_suggestions:
                    _items.append(_item_video_suggestions.to_dict())
            _dict['VideoSuggestions'] = _items
        # set to None if responsive_ad (nullable) is None
        # and model_fields_set contains the field
        if self.responsive_ad is None and "responsive_ad" in self.model_fields_set:
            _dict['ResponsiveAd'] = None

        # set to None if image_suggestions (nullable) is None
        # and model_fields_set contains the field
        if self.image_suggestions is None and "image_suggestions" in self.model_fields_set:
            _dict['ImageSuggestions'] = None

        # set to None if video_suggestions (nullable) is None
        # and model_fields_set contains the field
        if self.video_suggestions is None and "video_suggestions" in self.model_fields_set:
            _dict['VideoSuggestions'] = None

        # set to None if prompt_brand_warning (nullable) is None
        # and model_fields_set contains the field
        if self.prompt_brand_warning is None and "prompt_brand_warning" in self.model_fields_set:
            _dict['PromptBrandWarning'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of CreateResponsiveAdRecommendationResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "ResponsiveAd": ResponsiveAd.from_dict(obj["ResponsiveAd"]) if obj.get("ResponsiveAd") is not None else None,
                        "ImageSuggestions": [AdRecommendationImageSuggestion.from_dict(_item) for _item in obj["ImageSuggestions"]] if obj.get("ImageSuggestions") is not None else None,
                        "VideoSuggestions": [AdRecommendationVideoSuggestion.from_dict(_item) for _item in obj["VideoSuggestions"]] if obj.get("VideoSuggestions") is not None else None,
                        "PromptBrandWarning": obj.get("PromptBrandWarning") if obj.get("PromptBrandWarning") is not None else None
        })
        return _obj



from bingads.v13.internal.bulk.string_table import _StringTable
from bingads.v13.internal.bulk.entities.single_record_bulk_entity import _SingleRecordBulkEntity
from bingads.v13.internal.bulk.mappings import _SimpleBulkMapping
from bingads.v13.internal.extensions import *
from decimal import Decimal


class BulkImage(_SingleRecordBulkEntity):
    """ Represents a image that can be read or written in a bulk file.

    Properties of this class and of classes that it is derived from, correspond to fields of the Image record in a bulk file.
    For more information, see Image at https://go.microsoft.com/fwlink/?linkid=846127

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self, status = None):
        super(BulkImage, self).__init__()
        self._id = None
        self._status = status
        self._account_id = None
        self._text = None
        self._subtype = None
        self._url = None
        self._width = None
        self._height = None
        

    @property
    def id(self):
        """
        the Id object, see more detail at: https://go.microsoft.com/fwlink/?linkid=846127
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def status(self):
        """ the status of bulk record
        Corresponds to the 'Status' field in the bulk file.

        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    @property
    def account_id(self):
        """ the id of the account which contains the image
        Corresponds to the 'Parent Id' field in the bulk file.

        :rtype: long
        """
        return self._account_id
    
    @account_id.setter
    def account_id(self, value):
        self._account_id = value

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
       
    @property
    def subtype(self):
        return self._subtype 
    
    @subtype.setter
    def subtype(self, value):
        self._subtype = value

    @property
    def url(self):
        return self._url
     
    @url.setter
    def url(self, value):
        self._url = value

    @property
    def width(self):
        return self._width 

    @width.setter
    def width(self, value):
        self._width = value

    @property
    def height(self):
        return self._height    
  
    @height.setter
    def height(self, value):
        self._height = value

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.Id,
            field_to_csv=lambda c: bulk_str(c.id),
            csv_to_field=lambda c, v: setattr(c, 'id', int(v) if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.ParentId,
            field_to_csv=lambda c: bulk_str(c.account_id),
            csv_to_field=lambda c, v: setattr(c, 'account_id', int(v) if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.Status,
            field_to_csv=lambda c: c.status,
            csv_to_field=lambda c, v: setattr(c, 'status', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.SubType,
            field_to_csv=lambda c: bulk_str(c.subtype),
            csv_to_field=lambda c, v: setattr(c, 'subtype', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.Url,
            field_to_csv=lambda c: bulk_str(c.url),
            csv_to_field=lambda c, v: setattr(c, 'url', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.Text,
            field_to_csv=lambda c: bulk_str(c.text),
            csv_to_field=lambda c, v: setattr(c, 'text', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.Width,
            field_to_csv=lambda c: bulk_str(c.width),
            csv_to_field=lambda c, v: setattr(c, 'width', int(v) if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.Height,
            field_to_csv=lambda c: bulk_str(c.height),
            csv_to_field=lambda c, v: setattr(c, 'height', int(v) if v else None)
        ),
    ]

    def process_mappings_from_row_values(self, row_values):
        row_values.convert_to_entity(self, BulkImage._MAPPINGS)

    def process_mappings_to_row_values(self, row_values, exclude_readonly_data):
        self.convert_to_values(row_values, BulkImage._MAPPINGS)

    def read_additional_data(self, stream_reader):
        super(BulkImage, self).read_additional_data(stream_reader)



from bingads.service_client import _CAMPAIGN_OBJECT_FACTORY_V13

from bingads.v13.internal.bulk.entities.bulk_entity_identifier import _BulkEntityIdentifier
from bingads.v13.internal.bulk.string_table import _StringTable
from bingads.v13.internal.bulk.entities.single_record_bulk_entity import _SingleRecordBulkEntity
from bingads.v13.internal.bulk.mappings import _SimpleBulkMapping, _DynamicColumnNameMapping
from bingads.v13.internal.bulk.entities.multi_record_bulk_entity import _MultiRecordBulkEntity
from bingads.v13.internal.extensions import bulk_str


class _BulkNegativeSite(_SingleRecordBulkEntity):
    """ This abstract base class for the bulk negative sites that are assigned individually to a campaign or ad group entity.

    *See also:*

    * :class:`.BulkAdGroupNegativeSite`
    * :class:`.BulkCampaignNegativeSite`
    """

    def __init__(self, identifier, website=None):
        super(_BulkNegativeSite, self).__init__()

        self._identifier = identifier
        self._website = website

    @property
    def website(self):
        """ The URL of a website on which you do not want your ads displayed.

        Corresponds to the 'Website' field in the bulk file.

        :rtype: str
        """

        return self._website

    @website.setter
    def website(self, website):
        self._website = website

    @property
    def status(self):
        """ The status of the negative site association.

        :rtype: str
        """

        return self._identifier.status

    @status.setter
    def status(self, value):
        self._identifier.status = value

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.Website,
            field_to_csv=lambda c: c.website,
            csv_to_field=lambda c, v: setattr(c, 'website', v)
        ),
    ]

    def process_mappings_from_row_values(self, row_values):
        self._identifier.read_from_row_values(row_values)
        row_values.convert_to_entity(self, _BulkNegativeSite._MAPPINGS)

    def process_mappings_to_row_values(self, row_values, exclude_readonly_data):
        self._identifier.write_to_row_values(row_values, exclude_readonly_data)
        self.convert_to_values(row_values, _BulkNegativeSite._MAPPINGS)

    @property
    def can_enclose_in_multiline_entity(self):
        return True

    def enclose_in_multiline_entity(self):
        return self.create_negative_sites_with_this_negative_site()

    def create_negative_sites_with_this_negative_site(self):
        raise NotImplementedError()

    def read_additional_data(self, stream_reader):
        super(_BulkNegativeSite, self).read_additional_data(stream_reader)


class BulkAdGroupNegativeSite(_BulkNegativeSite):
    """ Represents a negative site that is assigned to an ad group. Each negative site can be read or written in a bulk file.

    This class exposes properties that can be read and written as fields of the Ad Group Negative Site record in a bulk file.

    For more information, see Ad Group Negative Site at https://go.microsoft.com/fwlink/?linkid=846127.

    One :class:`.BulkAdGroupNegativeSites` exposes a read only list of :class:`.BulkAdGroupNegativeSite`. Each
    :class:`.BulkAdGroupNegativeSite` instance corresponds to one Ad Group Negative Site record in the bulk file. If you
    upload a :class:`.BulkAdGroupNegativeSites`, then you are effectively replacing any existing negative sites
    assigned to the ad group.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self,
                 status=None,
                 ad_group_id=None,
                 ad_group_name=None,
                 campaign_name=None,
                 website=None):
        super(BulkAdGroupNegativeSite, self).__init__(
            _BulkAdGroupNegativeSitesIdentifier(
                status=status,
                ad_group_id=ad_group_id,
                ad_group_name=ad_group_name,
                campaign_name=campaign_name,
            ),
            website=website
        )

    @property
    def ad_group_id(self):
        """ The identifier of the ad group that the negative site is assigned.

        Corresponds to the 'Parent Id' field in the bulk file.

        :rtype: int
        """

        return self._identifier.ad_group_id

    @ad_group_id.setter
    def ad_group_id(self, value):
        self._identifier.ad_group_id = value

    @property
    def ad_group_name(self):
        """ The name of the ad group that the negative site is assigned.

        Corresponds to the 'Ad Group' field in the bulk file.

        :rtype: str
        """

        return self._identifier.ad_group_name

    @ad_group_name.setter
    def ad_group_name(self, value):
        self._identifier.ad_group_name = value

    @property
    def campaign_name(self):
        """ The name of the ad group that the negative site is assigned.

        Corresponds to the 'Ad Group' field in the bulk file.

        :rtype: str
        """

        return self._identifier.campaign_name

    @campaign_name.setter
    def campaign_name(self, value):
        self._identifier.campaign_name = value

    def create_negative_sites_with_this_negative_site(self):
        return BulkAdGroupNegativeSites(site=self)


class BulkCampaignNegativeSite(_BulkNegativeSite):
    """ Represents a negative site that is assigned to an campaign. Each negative site can be read or written in a bulk file.

    This class exposes properties that can be read and written as fields of the Campaign Negative Site record in a bulk file.

    For more information, see Campaign Negative Site at https://go.microsoft.com/fwlink/?linkid=846127.

    One :class:`.BulkCampaignNegativeSites` exposes a read only list of :class:`.BulkCampaignNegativeSite`. Each
    :class:`.BulkCampaignNegativeSite` instance corresponds to one Campaign Negative Site record in the bulk file. If you
    upload a :class:`.BulkCampaignNegativeSites`, then you are effectively replacing any existing negative sites
    assigned to the campaign.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self,
                 status=None,
                 campaign_id=None,
                 campaign_name=None,
                 website=None):
        super(BulkCampaignNegativeSite, self).__init__(
            _BulkCampaignNegativeSitesIdentifier(
                status=status,
                campaign_id=campaign_id,
                campaign_name=campaign_name
            ),
            website=website
        )

    @property
    def campaign_id(self):
        """ The identifier of the campaign that the negative site is assigned.

        Corresponds to the 'Parent Id' field in the bulk file.

        :rtype: int
        """

        return self._identifier.campaign_id

    @campaign_id.setter
    def campaign_id(self, value):
        self._identifier.campaign_id = value

    @property
    def campaign_name(self):
        """ The name of the campaign that the negative site is assigned.

        Corresponds to the 'Campaign' field in the bulk file.

        :rtype: str
        """
        return self._identifier.campaign_name

    @campaign_name.setter
    def campaign_name(self, value):
        self._identifier.campaign_name = value

    def create_negative_sites_with_this_negative_site(self):
        return BulkCampaignNegativeSites(site=self)


class _BulkNegativeSites(_MultiRecordBulkEntity):
    """ This abstract base class for the bulk negative sites that assigned in sets to a campaign or ad group entity. """

    def __init__(self, status=None, site=None, identifier=None):
        super(_BulkNegativeSites, self).__init__()

        self._bulk_negative_sites = []
        self._first_row_identifier = None
        self._has_delete_all_row = None

        self._site = site
        self._identifier = identifier

        if self._site and self._identifier:
            raise ValueError('Conflicting keyword arguments of site and identifier provided')

        if self._site:
            if not isinstance(self._site, self.site_class):
                raise ValueError('Negative site object provided is not of type: {0}'.format(self.site_class.__name__))
            self._bulk_negative_sites.append(self._site)
            self._identifier = self._site._identifier

        if self._identifier:
            if not isinstance(self._identifier, self.identifier_class):
                raise ValueError(
                    'Negative site object provided is not of type: {0}'.format(self.identifier_class.__name__))
            self._first_row_identifier = self._identifier
            self._has_delete_all_row = self._identifier.is_delete_row

        self._status = status

    @property
    def status(self):
        """ The status of the negative site association.

        :rtype: str
        """

        return self._status

    @status.setter
    def status(self, status):
        self._status = status

    @property
    def negative_sites(self):
        return self._bulk_negative_sites

    @property
    def child_entities(self):
        return self.negative_sites

    def _create_identifier(self):
        raise NotImplementedError()

    def _validate_properties_not_null(self):
        raise NotImplementedError()

    def write_to_stream(self, row_writer, exclude_readonly_data):
        self._validate_properties_not_null()

        delete_row = self._create_identifier()
        delete_row._status = 'Deleted'
        row_writer.write_object_row(delete_row, exclude_readonly_data)

        if self._status == 'Deleted':
            return

        for site in self.convert_api_to_bulk_negative_sites():
            site.write_to_stream(row_writer, exclude_readonly_data)

    def convert_api_to_bulk_negative_sites(self):
        raise NotImplementedError()

    def reconstruct_api_objects(self):
        raise NotImplementedError()

    @property
    def site_class(self):
        raise NotImplementedError()

    @property
    def identifier_class(self):
        raise NotImplementedError()

    def read_related_data_from_stream(self, stream_reader):
        has_more_rows = True
        while has_more_rows:
            site_success, site = stream_reader.try_read(
                self.site_class,
                lambda x: x._identifier == self._first_row_identifier
            )
            if site_success:
                self._bulk_negative_sites.append(site)
            else:
                identifier_success, identifier = stream_reader.try_read(
                    self.identifier_class,
                    lambda x: x == self._first_row_identifier
                )
                if identifier_success:
                    if identifier.is_delete_row:
                        self._has_delete_all_row = True
                else:
                    has_more_rows = False

        self.reconstruct_api_objects()
        self._status = 'Active' if self._bulk_negative_sites else 'Deleted'

    @property
    def all_children_are_present(self):
        return self._has_delete_all_row


class BulkAdGroupNegativeSites(_BulkNegativeSites):
    """ Represents one or more negative sites that are assigned to an ad group. Each negative site can be read or written in a bulk file.

    This class exposes properties that can be read and written as fields of the Ad Group Negative Site record in a bulk file.

    For more information, see Ad Group Negative Site at https://go.microsoft.com/fwlink/?linkid=846127.

    One :class:`.BulkAdGroupNegativeSites` has one or more :class:`.BulkAdGroupNegativeSite`. Each :class:`.BulkAdGroupNegativeSite` instance
    corresponds to one Ad Group Negative Site record in the bulk file. If you upload a :class:`.BulkAdGroupNegativeSites`,
    then you are effectively replacing any existing negative sites assigned to the ad group.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self,
                 ad_group_negative_sites=None,
                 ad_group_name=None,
                 campaign_name=None,
                 status=None,
                 site=None,
                 identifier=None):
        super(BulkAdGroupNegativeSites, self).__init__(
            status=status,
            site=site,
            identifier=identifier,
        )

        self._ad_group_negative_sites = ad_group_negative_sites
        self._ad_group_name = ad_group_name
        self._campaign_name = campaign_name

        if self._identifier:
            self.set_data_from_identifier(self._identifier)

    @property
    def ad_group_negative_sites(self):
        """ The AdGroupNegativeSites Data Object of the Campaign Management Service.

        subset of AdGroupNegativeSites properties are available in the Ad Group Negative Site record.
        For more information, see Ad Group Negative Site at https://go.microsoft.com/fwlink/?linkid=846127.
        """

        return self._ad_group_negative_sites

    @ad_group_negative_sites.setter
    def ad_group_negative_sites(self, ad_group_negative_sites):
        self._ad_group_negative_sites = ad_group_negative_sites

    @property
    def ad_group_name(self):
        """ The name of the ad group that the negative site is assigned.

        Corresponds to the 'Ad Group' field in the bulk file.

        :rtype: str
        """

        return self._ad_group_name

    @ad_group_name.setter
    def ad_group_name(self, ad_group_name):
        self._ad_group_name = ad_group_name

    @property
    def campaign_name(self):
        """ The name of the campaign that the negative site is assigned.

        Corresponds to the 'Campaign' field in the bulk file.

        :rtype: str
        """

        return self._campaign_name

    @campaign_name.setter
    def campaign_name(self, campaign_name):
        self._campaign_name = campaign_name

    def set_data_from_identifier(self, identifier):
        self._ad_group_negative_sites = _CAMPAIGN_OBJECT_FACTORY_V13.create('AdGroupNegativeSites')
        self._ad_group_negative_sites.AdGroupId = identifier.ad_group_id
        self._ad_group_name = identifier.ad_group_name
        self._campaign_name = identifier.campaign_name

    def convert_api_to_bulk_negative_sites(self):
        self._validate_list_not_null_or_empty(
            self._ad_group_negative_sites.NegativeSites,
            self._ad_group_negative_sites.NegativeSites.string,
            'ad_group_negative_sites.negative_sites'
        )

        def convert_api_to_bulk_negative_site(website):
            bulk_ad_group_negative_site = BulkAdGroupNegativeSite()
            bulk_ad_group_negative_site.ad_group_id = self._ad_group_negative_sites.AdGroupId
            bulk_ad_group_negative_site.ad_group_name = self._ad_group_name
            bulk_ad_group_negative_site.campaign_name = self._campaign_name
            bulk_ad_group_negative_site._website = website
            return bulk_ad_group_negative_site

        return map(convert_api_to_bulk_negative_site, self._ad_group_negative_sites.NegativeSites.string)

    def reconstruct_api_objects(self):
        self._ad_group_negative_sites.NegativeSites.string = list(map(lambda x: x.website, self.negative_sites))

    def _create_identifier(self):
        return _BulkAdGroupNegativeSitesIdentifier(
            ad_group_id=self._ad_group_negative_sites.AdGroupId,
            ad_group_name=self._ad_group_name,
            campaign_name=self.campaign_name
        )

    def _validate_properties_not_null(self):
        self._validate_property_not_null(self._ad_group_negative_sites, 'ad_group_negative_sites')

    @property
    def identifier_class(self):
        return _BulkAdGroupNegativeSitesIdentifier

    @property
    def site_class(self):
        return BulkAdGroupNegativeSite


class BulkCampaignNegativeSites(_BulkNegativeSites):
    """ Represents one or more negative sites that are assigned to an campaign. Each negative site can be read or written in a bulk file.

    This class exposes properties that can be read and written as fields of the Campaign Negative Site record in a bulk file.

    For more information, see Campaign Negative Site at https://go.microsoft.com/fwlink/?linkid=846127.

    One :class:`.BulkCampaignNegativeSites` has one or more :class:`.BulkCampaignNegativeSite`. Each :class:`.BulkCampaignNegativeSite` instance
    corresponds to one Campaign Negative Site record in the bulk file. If you upload a :class:`.BulkCampaignNegativeSites`,
    then you are effectively replacing any existing negative sites assigned to the campaign.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self,
                 campaign_negative_sites=None,
                 campaign_name=None,
                 status=None,
                 site=None,
                 identifier=None):
        super(BulkCampaignNegativeSites, self).__init__(
            status=status,
            site=site,
            identifier=identifier
        )

        self._campaign_negative_sites = campaign_negative_sites
        self._campaign_name = campaign_name

        if self._identifier:
            self.set_data_from_identifier(self._identifier)

    @property
    def campaign_negative_sites(self):
        """ The CampaignNegativeSites Data Object of the Campaign Management Service.

        A subset of CampaignNegativeSites properties are available in the Campaign Negative Site record.
        For more information, see Campaign Negative Site at https://go.microsoft.com/fwlink/?linkid=846127.

        """

        return self._campaign_negative_sites

    @campaign_negative_sites.setter
    def campaign_negative_sites(self, campaign_negative_sites):
        self._campaign_negative_sites = campaign_negative_sites

    @property
    def campaign_name(self):
        """ The name of the campaign that the negative site is assigned.

        Corresponds to the 'Campaign' field in the bulk file.
        :rtype: str
        """

        return self._campaign_name

    @campaign_name.setter
    def campaign_name(self, campaign_name):
        self._campaign_name = campaign_name

    def set_data_from_identifier(self, identifier):
        self._campaign_negative_sites = _CAMPAIGN_OBJECT_FACTORY_V13.create('CampaignNegativeSites')
        self.campaign_negative_sites.CampaignId = identifier.campaign_id
        self._campaign_name = identifier.campaign_name

    def convert_api_to_bulk_negative_sites(self):
        self._validate_list_not_null_or_empty(
            self._campaign_negative_sites.NegativeSites,
            self._campaign_negative_sites.NegativeSites.string,
            'campaign_negative_sites.negative_sites'
        )

        def convert_api_to_bulk_negative_site(website):
            bulk_campaign_negative_site = BulkCampaignNegativeSite()
            bulk_campaign_negative_site.campaign_id = self._campaign_negative_sites.CampaignId
            bulk_campaign_negative_site.campaign_name = self._campaign_name
            bulk_campaign_negative_site._website = website
            return bulk_campaign_negative_site

        return map(convert_api_to_bulk_negative_site, self._campaign_negative_sites.NegativeSites.string)

    def reconstruct_api_objects(self):
        self._campaign_negative_sites.NegativeSites.string = list(map(lambda x: x.website, self.negative_sites))

    def _create_identifier(self):
        return _BulkCampaignNegativeSitesIdentifier(
            campaign_id=self._campaign_negative_sites.CampaignId,
            campaign_name=self._campaign_name
        )

    def _validate_properties_not_null(self):
        self._validate_property_not_null(self._campaign_negative_sites, 'campaign_negative_sites')

    @property
    def identifier_class(self):
        return _BulkCampaignNegativeSitesIdentifier

    @property
    def site_class(self):
        return BulkCampaignNegativeSite


class _BulkNegativeSiteIdentifier(_BulkEntityIdentifier):
    def __init__(self, status=None, entity_id=None, entity_name=None):
        self._status = status
        self._entity_id = entity_id
        self._entity_name = entity_name

    @property
    def status(self):
        return self._status

    @property
    def entity_id(self):
        return self._entity_id

    @property
    def entity_name(self):
        return self._entity_name

    @property
    def _parent_column_name(self):
        raise NotImplementedError()

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.Status,
            field_to_csv=lambda c: bulk_str(c._status),
            csv_to_field=lambda c, v: setattr(c, '_status', v if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.ParentId,
            field_to_csv=lambda c: None if c._entity_id == 0 else bulk_str(c._entity_id),
            csv_to_field=lambda c, v: setattr(c, '_entity_id', int(v) if v else 0)
        ),
        _DynamicColumnNameMapping(
            header_func=lambda c: c._parent_column_name,
            field_to_csv=lambda c: c._entity_name,
            csv_to_field=lambda c, v: setattr(c, '_entity_name', v)
        )
    ]

    @property
    def is_delete_row(self):
        return self._status == 'Deleted'

    def read_from_row_values(self, row_values):
        row_values.convert_to_entity(self, _BulkNegativeSiteIdentifier._MAPPINGS)

    def write_to_row_values(self, row_values, exclude_readonly_data):
        self.convert_to_values(row_values, _BulkNegativeSiteIdentifier._MAPPINGS)


class _BulkCampaignNegativeSitesIdentifier(_BulkNegativeSiteIdentifier):
    def __init__(self, status=None, campaign_id=None, campaign_name=None):
        super(_BulkCampaignNegativeSitesIdentifier, self).__init__(
            status,
            campaign_id,
            campaign_name,
        )

    def __eq__(self, other):
        is_name_not_empty = (
            self.campaign_name is not None and
            len(self.campaign_name) > 0
        )
        return (
            type(self) == type(other) and
            (
                self.campaign_id == other.campaign_id or
                (
                    is_name_not_empty and
                    self.campaign_name == other.campaign_name
                )
            )
        )

    @property
    def campaign_id(self):
        return self._entity_id

    @campaign_id.setter
    def campaign_id(self, value):
        self._entity_id = value

    @property
    def campaign_name(self):
        return self._entity_name

    @campaign_name.setter
    def campaign_name(self, value):
        self._entity_name = value

    def _create_entity_with_this_identifier(self):
        return BulkCampaignNegativeSites(identifier=self)

    @property
    def _parent_column_name(self):
        return _StringTable.Campaign


class _BulkAdGroupNegativeSitesIdentifier(_BulkNegativeSiteIdentifier):
    def __init__(self,
                 status=None,
                 ad_group_id=None,
                 ad_group_name=None,
                 campaign_name=None):
        super(_BulkAdGroupNegativeSitesIdentifier, self).__init__(
            status,
            ad_group_id,
            ad_group_name,
        )
        self._campaign_name = campaign_name

    def __eq__(self, other):
        is_name_not_empty = (
            self.campaign_name is not None and
            len(self.campaign_name) > 0 and
            self.ad_group_name is not None and
            len(self.ad_group_name) > 0
        )
        return (
            type(self) == type(other) and
            (
                self.ad_group_id == other.ad_group_id or
                (
                    is_name_not_empty and
                    self.campaign_name == other.campaign_name and
                    self.ad_group_name == other.ad_group_name
                )
            )
        )

    @property
    def ad_group_id(self):
        return self._entity_id

    @ad_group_id.setter
    def ad_group_id(self, value):
        self._entity_id = value

    @property
    def ad_group_name(self):
        return self._entity_name

    @ad_group_name.setter
    def ad_group_name(self, value):
        self._entity_name = value

    @property
    def campaign_name(self):
        return self._campaign_name

    @campaign_name.setter
    def campaign_name(self, campaign_name):
        self._campaign_name = campaign_name

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.Campaign,
            field_to_csv=lambda c: c.campaign_name,
            csv_to_field=lambda c, v: setattr(c, 'campaign_name', v)
        ),
    ]

    def read_from_row_values(self, row_values):
        super(_BulkAdGroupNegativeSitesIdentifier, self).read_from_row_values(row_values)
        row_values.convert_to_entity(self, _BulkAdGroupNegativeSitesIdentifier._MAPPINGS)

    def write_to_row_values(self, row_values, exclude_readonly_data):
        super(_BulkAdGroupNegativeSitesIdentifier, self).write_to_row_values(row_values, exclude_readonly_data)
        self.convert_to_values(row_values, _BulkAdGroupNegativeSitesIdentifier._MAPPINGS)

    def _create_entity_with_this_identifier(self):
        return BulkAdGroupNegativeSites(identifier=self)

    @property
    def _parent_column_name(self):
        return _StringTable.AdGroup



# coding: utf-8

"""
    ReportingService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self

class AssetPerformanceReportColumn(str, Enum):
    TIMEPERIOD = 'TimePeriod'

    ACCOUNTID = 'AccountId'

    ACCOUNTNAME = 'AccountName'

    CAMPAIGNID = 'CampaignId'

    CAMPAIGNNAME = 'CampaignName'

    ADGROUPID = 'AdGroupId'

    ADGROUPNAME = 'AdGroupName'

    ASSETID = 'AssetId'

    ASSETCONTENT = 'AssetContent'

    ASSETTYPE = 'AssetType'

    ASSETSOURCE = 'AssetSource'

    IMPRESSIONS = 'Impressions'

    CLICKS = 'Clicks'

    CTR = 'Ctr'

    SPEND = 'Spend'

    VIDEOVIEWS = 'VideoViews'

    VIDEOVIEWSAT25PERCENT = 'VideoViewsAt25Percent'

    VIDEOVIEWSAT50PERCENT = 'VideoViewsAt50Percent'

    VIDEOVIEWSAT75PERCENT = 'VideoViewsAt75Percent'

    COMPLETEDVIDEOVIEWS = 'CompletedVideoViews'

    VIDEOCOMPLETIONRATE = 'VideoCompletionRate'

    CONVERSIONS = 'Conversions'

    REVENUE = 'Revenue'

	
    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of AssetPerformanceReportColumn from a JSON string"""
        return cls(json.loads(json_str))





from bingads.v13.internal.bulk.mappings import _SimpleBulkMapping
from bingads.v13.internal.bulk.string_table import _StringTable
from bingads.service_client import _CAMPAIGN_OBJECT_FACTORY_V13

from .common import *
from .common import _BulkAdExtensionBase
from .common import _BulkCampaignAdExtensionAssociation

_CallAdExtension = type(_CAMPAIGN_OBJECT_FACTORY_V13.create('CallAdExtension'))


class BulkCallAdExtension(_BulkAdExtensionBase):
    """ Represents a call ad extension.

    This class exposes the :attr:`call_ad_extension` property that can be read and written
    as fields of the Call Ad Extension record in a bulk file.

    For more information, see Call Ad Extension at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self,
                 account_id=None,
                 ad_extension=None):
        if ad_extension and not isinstance(ad_extension, _CallAdExtension):
            raise ValueError('The type of ad_extension is: {0}, should be: {1}'.format(
                type(ad_extension),
                'CallAdExtension'
            ))
        super(BulkCallAdExtension, self).__init__(
            account_id=account_id,
            ad_extension=ad_extension
        )

    @property
    def call_ad_extension(self):
        """ The call ad extension.

        see Call Ad Extension at https://go.microsoft.com/fwlink/?linkid=846127.
        """

        return self._ad_extension

    @call_ad_extension.setter
    def call_ad_extension(self, value):
        self._ad_extension = value

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.PhoneNumber,
            field_to_csv=lambda c: c.call_ad_extension.PhoneNumber,
            csv_to_field=lambda c, v: setattr(c.call_ad_extension, 'PhoneNumber', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.CountryCode,
            field_to_csv=lambda c: c.call_ad_extension.CountryCode,
            csv_to_field=lambda c, v: setattr(c.call_ad_extension, 'CountryCode', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.IsCallOnly,
            field_to_csv=lambda c: bulk_str(c.call_ad_extension.IsCallOnly),
            csv_to_field=lambda c, v: setattr(c.call_ad_extension, 'IsCallOnly',
                                              v.lower() == 'true' if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.IsCallTrackingEnabled,
            field_to_csv=lambda c: bulk_str(c.call_ad_extension.IsCallTrackingEnabled),
            csv_to_field=lambda c, v: setattr(c.call_ad_extension, 'IsCallTrackingEnabled',
                                              v.lower() == 'true' if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.RequireTollFreeTrackingNumber,
            field_to_csv=lambda c: bulk_str(c.call_ad_extension.RequireTollFreeTrackingNumber),
            csv_to_field=lambda c, v: setattr(c.call_ad_extension, 'RequireTollFreeTrackingNumber',
                                              v.lower() == 'true' if v else None)
        ),
    ]

    def process_mappings_from_row_values(self, row_values):
        self.call_ad_extension = _CAMPAIGN_OBJECT_FACTORY_V13.create('CallAdExtension')
        self.call_ad_extension.Type = 'CallAdExtension'
        super(BulkCallAdExtension, self).process_mappings_from_row_values(row_values)
        row_values.convert_to_entity(self, BulkCallAdExtension._MAPPINGS)

    def process_mappings_to_row_values(self, row_values, exclude_readonly_data):
        self._validate_property_not_null(self.call_ad_extension, 'call_ad_extension')
        super(BulkCallAdExtension, self).process_mappings_to_row_values(row_values, exclude_readonly_data)
        self.convert_to_values(row_values, BulkCallAdExtension._MAPPINGS)


class BulkCampaignCallAdExtension(_BulkCampaignAdExtensionAssociation):
    """ Represents a campaign level call ad extension.

    This class exposes properties that can be read and written
    as fields of the Campaign Call Ad Extension record in a bulk file.

    For more information, see Campaign Call Ad Extension at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    pass



# coding: utf-8

"""
    AdInsightService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union, Set
from openapi_client.models.adinsight.ad_group_estimator import AdGroupEstimator
from openapi_client.models.adinsight.criterion import Criterion
from openapi_client.models.adinsight.negative_keyword import NegativeKeyword
from typing_extensions import Self

class CampaignEstimator(BaseModel):
    """
    CampaignEstimator
    """ # noqa: E501
    campaign_id: Optional[StrictStr] = Field(default=None, alias="CampaignId")
    daily_budget: Optional[StrictFloat] = Field(default=None, alias="DailyBudget")
    criteria: Optional[List[Optional[Criterion]]] = Field(default=None, alias="Criteria")
    ad_group_estimators: Optional[List[Optional[AdGroupEstimator]]] = Field(default=None, alias="AdGroupEstimators")
    negative_keywords: Optional[List[Optional[NegativeKeyword]]] = Field(default=None, alias="NegativeKeywords")
    __properties: ClassVar[List[str]] = ["CampaignId", "DailyBudget", "Criteria", "AdGroupEstimators", "NegativeKeywords"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )
	

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in criteria (list)
        _items = []
        if self.criteria:
            for _item_criteria in self.criteria:
                if _item_criteria:
                    _items.append(_item_criteria.to_dict())
            _dict['Criteria'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in ad_group_estimators (list)
        _items = []
        if self.ad_group_estimators:
            for _item_ad_group_estimators in self.ad_group_estimators:
                if _item_ad_group_estimators:
                    _items.append(_item_ad_group_estimators.to_dict())
            _dict['AdGroupEstimators'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in negative_keywords (list)
        _items = []
        if self.negative_keywords:
            for _item_negative_keywords in self.negative_keywords:
                if _item_negative_keywords:
                    _items.append(_item_negative_keywords.to_dict())
            _dict['NegativeKeywords'] = _items
        # set to None if campaign_id (nullable) is None
        # and model_fields_set contains the field
        if self.campaign_id is None and "campaign_id" in self.model_fields_set:
            _dict['CampaignId'] = None

        # set to None if daily_budget (nullable) is None
        # and model_fields_set contains the field
        if self.daily_budget is None and "daily_budget" in self.model_fields_set:
            _dict['DailyBudget'] = None

        # set to None if criteria (nullable) is None
        # and model_fields_set contains the field
        if self.criteria is None and "criteria" in self.model_fields_set:
            _dict['Criteria'] = None

        # set to None if ad_group_estimators (nullable) is None
        # and model_fields_set contains the field
        if self.ad_group_estimators is None and "ad_group_estimators" in self.model_fields_set:
            _dict['AdGroupEstimators'] = None

        # set to None if negative_keywords (nullable) is None
        # and model_fields_set contains the field
        if self.negative_keywords is None and "negative_keywords" in self.model_fields_set:
            _dict['NegativeKeywords'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of CampaignEstimator from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "CampaignId": obj.get("CampaignId") if obj.get("CampaignId") is not None else None,
                        "DailyBudget": obj.get("DailyBudget") if obj.get("DailyBudget") is not None else None,
                        "Criteria": [Criterion.from_dict(_item) for _item in obj["Criteria"]] if obj.get("Criteria") is not None else None,
                        "AdGroupEstimators": [AdGroupEstimator.from_dict(_item) for _item in obj["AdGroupEstimators"]] if obj.get("AdGroupEstimators") is not None else None,
                        "NegativeKeywords": [NegativeKeyword.from_dict(_item) for _item in obj["NegativeKeywords"]] if obj.get("NegativeKeywords") is not None else None
        })
        return _obj



class BulkOperationProgressInfo(object):
    """ Contains percent complete progress information for the bulk operation."""

    def __init__(self, percent_complete=0):
        """ Initialize a new instance of this class.

        :param percent_complete: (optional) Percent complete progress information for the bulk operation.
        :type percent_complete: int
        """

        self._percent_complete = percent_complete

    @property
    def percent_complete(self):
        """ Percent complete progress information for the bulk operation.

        :rtype: int
        """

        return self._percent_complete



from bingads.v13.bulk.entities import *
from bingads.service_client import _CAMPAIGN_OBJECT_FACTORY_V13
from bingads.v13.bulk.entities.target_criterions.bulk_campaign_biddable_criterion import BulkCampaignBiddableCriterion
from bingads.v13.internal.bulk.mappings import _SimpleBulkMapping
from bingads.v13.internal.bulk.string_table import _StringTable
from bingads.v13.internal.extensions import *

class BulkCampaignDeviceCriterion(BulkCampaignBiddableCriterion):
    """ Represents an Campaign Device Criterion that can be read or written in a bulk file.

    This class exposes the :attr:`biddable_campaign_criterion` property that can be read and written as fields of the
    Campaign Device Criterion record in a bulk file.

    For more information, see Campaign Device Criterion at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self,
                 biddable_campaign_criterion=None,
                 campaign_name=None, ):
        super(BulkCampaignDeviceCriterion, self).__init__(biddable_campaign_criterion, campaign_name)

    _MAPPINGS = [
        _SimpleBulkMapping(
            _StringTable.Target,
            field_to_csv=lambda c: field_to_csv_DeviceTarget(c.biddable_campaign_criterion),
            csv_to_field=lambda c, v: csv_to_field_DeviceTarget(c.biddable_campaign_criterion, v)
        ),
        _SimpleBulkMapping(
            _StringTable.OsNames,
            field_to_csv=lambda c: field_to_csv_OSName(c.biddable_campaign_criterion),
            csv_to_field=lambda c, v: csv_to_field_OSName(c.biddable_campaign_criterion, v)
        ),
    ]

    def create_criterion(self):
        self._biddable_campaign_criterion.Criterion = _CAMPAIGN_OBJECT_FACTORY_V13.create('DeviceCriterion')
        self._biddable_campaign_criterion.Criterion.Type = 'DeviceCriterion'
        
    def process_mappings_to_row_values(self, row_values, exclude_readonly_data):
        super(BulkCampaignDeviceCriterion, self).process_mappings_to_row_values(row_values, exclude_readonly_data)
        self.convert_to_values(row_values, BulkCampaignDeviceCriterion._MAPPINGS)

    def process_mappings_from_row_values(self, row_values):
        super(BulkCampaignDeviceCriterion, self).process_mappings_from_row_values(row_values)
        row_values.convert_to_entity(self, BulkCampaignDeviceCriterion._MAPPINGS)

    def read_additional_data(self, stream_reader):
        super(BulkCampaignDeviceCriterion, self).read_additional_data(stream_reader)



# coding: utf-8

# flake8: noqa

"""
    CampaignMTServiceHost | v1

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


__version__ = "1.0.0"

# import apis into sdk package
from openapi_client.api.bulk_service_api import BulkServiceApi
from openapi_client.api.campaign_management_service_api import CampaignManagementServiceApi

# import ApiClient
from openapi_client.api_response import ApiResponse
from openapi_client.api_client import ApiClient
from openapi_client.configuration import Configuration
from openapi_client.exceptions import OpenApiException
from openapi_client.exceptions import ApiTypeError
from openapi_client.exceptions import ApiValueError
from openapi_client.exceptions import ApiKeyError
from openapi_client.exceptions import ApiAttributeError
from openapi_client.exceptions import ApiException

# import models into sdk package
from openapi_client.models.campaign import *
from openapi_client.models.bulk import *
from openapi_client.models.reporting import *



from abc import ABCMeta, abstractproperty
from bingads.v13.bulk.entities.bulk_entity import BulkEntity


class _MultiRecordBulkEntity(BulkEntity, metaclass=ABCMeta):
    """ Bulk entity that has its data in multiple records within the bulk file.

    For more information, see Bulk File Schema at https://go.microsoft.com/fwlink/?linkid=846127.
    """

    def __init__(self):
        super(_MultiRecordBulkEntity, self).__init__()

    @abstractproperty
    def child_entities(self):
        """ The child entities that this multi record entity contains.

        :rtype: list[BulkEntity]
        """

        raise NotImplementedError()

    @abstractproperty
    def all_children_are_present(self):
        """ True, if the object is fully constructed (contains all of its children), determined by the presence of delete all row, False otherwise

        :rtype: bool
        """

        raise NotImplementedError()

    @property
    def has_errors(self):
        """ Indicates whether or not the errors property of any of the ChildEntities is null or empty.

        If true, one or more ChildEntities contains the details of one or more :class:`.BulkError` objects.

        :rtype: bool
        """

        return any(map(lambda x: x.has_errors, self.child_entities))

    @property
    def last_modified_time(self):
        """ Gets the last modified time for the first child entity, or null if there are no ChildEntities.

        :rtype: datetime.datetime
        """

        return self.child_entities[0].last_modified_time if self.child_entities else None

    @property
    def can_enclose_in_multiline_entity(self):
        return super(_MultiRecordBulkEntity, self).can_enclose_in_multiline_entity

    def enclose_in_multiline_entity(self):
        return super(_MultiRecordBulkEntity, self).enclose_in_multiline_entity()

    def read_from_row_values(self, row_values):
        super(_MultiRecordBulkEntity, self).read_from_row_values(row_values)

    def write_to_row_values(self, row_values):
        super(_MultiRecordBulkEntity, self).write_to_row_values(row_values)



import time
from bingads.exceptions import TimeoutException
from openapi_client.exceptions import ApiException

class _TimeHelper(object):
    """
    Time helper class to handle timeout
    """
    MIN_TIMEOUT_VALUE = 1000

    @staticmethod
    def get_current_time():
        return time.time()

    @staticmethod
    def get_current_time_milliseconds():
        return int(round(time.time()) * 1000)

    @staticmethod
    def get_remaining_time_milliseconds(start_time_milliseconds, total_timeout):
        return None if total_timeout is None else total_timeout - (_TimeHelper.get_current_time_milliseconds() - start_time_milliseconds)

    @staticmethod
    def get_remaining_time_milliseconds_with_min_value(start_time_milliseconds, total_timeout):
        if total_timeout is None:
            return None
        remaining_time = total_timeout - (_TimeHelper.get_current_time_milliseconds() - start_time_milliseconds)
        return remaining_time if remaining_time > 0 else _TimeHelper.MIN_TIMEOUT_VALUE


class _PollingBlocker(object):
    INITIAL_STATUS_CHECK_INTERVAL_IN_MS = 1000
    NUMBER_OF_INITIAL_STATUS_CHECKS = 5

    def __init__(self, poll_interval_in_milliseconds, timeout_in_milliseconds=None):
        self._poll_interval_in_milliseconds = poll_interval_in_milliseconds
        self._status_update_count = 0
        self._timeout_stamp = None if timeout_in_milliseconds is None else \
            int(round(time.time()) * 1000) + timeout_in_milliseconds

    def wait(self):
        self._status_update_count += 1
        if self._timeout_stamp is not None and int(round(time.time()) * 1000) > self._timeout_stamp:
            raise TimeoutException('Timeout at polling.')
        if self._status_update_count >= _PollingBlocker.NUMBER_OF_INITIAL_STATUS_CHECKS:
            time.sleep(self._poll_interval_in_milliseconds / 1000.0)
        else:
            time.sleep(_PollingBlocker.INITIAL_STATUS_CHECK_INTERVAL_IN_MS / 1000.0)


ratelimit_retry_duration=[15, 20, 25, 30]
def errorcode_of_exception(ex):
    if isinstance(ex, ApiException):
        try:
            body = ex.body
            if isinstance(body, dict) and 'Errors' in body:
                errors = body['Errors']
                if isinstance(errors, list) and errors:
                    return errors[0].get('Code', -1)
                return errors.get('Code', -1)
        except (AttributeError, KeyError):
            pass
    return -1


def operation_errorcode_of_exception(ex):
    if isinstance(ex, ApiException):
        try:
            body = ex.body
            if isinstance(body, dict) and 'OperationErrors' in body:
                operation_errors = body['OperationErrors']
                if isinstance(operation_errors, list) and operation_errors:
                    return operation_errors[0].get('ErrorCode', "")
                return operation_errors.get('ErrorCode', "")
        except (AttributeError, KeyError):
            pass
    return ""


from bingads.service_client import _CAMPAIGN_OBJECT_FACTORY_V13
from bingads.v13.internal.bulk.string_table import _StringTable
from bingads.v13.internal.bulk.entities.single_record_bulk_entity import _SingleRecordBulkEntity
from bingads.v13.internal.bulk.mappings import _SimpleBulkMapping
from bingads.v13.internal.extensions import *


class BulkOfflineConversion(_SingleRecordBulkEntity):
    """ Represents an offline conversion that can be read or written in a bulk file.

    This class exposes the :attr:`offline_conversion` property that can be read and written as fields of the Keyword record in a bulk file.
    Properties of this class and of classes that it is derived from, correspond to fields of the Keyword record in a bulk file.
    For more information, see Keyword at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self, offline_conversion=None):
        super(BulkOfflineConversion, self).__init__()
        self._offline_conversion = offline_conversion
        self._adjustment_value = None
        self._adjustment_time = None
        self._adjustment_type = None
        self._adjustment_currency_code = None
        self._external_attribution_model = None
        self._external_attribution_credit = None

    @property
    def adjustment_value(self):
        return self._adjustment_value;

    @adjustment_value.setter
    def adjustment_value(self, value):
        self._adjustment_value = value

    @property
    def adjustment_time(self):
        return self._adjustment_time;

    @adjustment_time.setter
    def adjustment_time(self, value):
        self._adjustment_time = value

    @property
    def adjustment_type(self):
        return self._adjustment_type;

    @adjustment_type.setter
    def adjustment_type(self, value):
        self._adjustment_type = value

    @property
    def adjustment_currency_code(self):
        return self._adjustment_currency_code;

    @adjustment_currency_code.setter
    def adjustment_currency_code(self, value):
        self._adjustment_currency_code = value

    @property
    def external_attribution_model(self):
        return self._external_attribution_model;

    @external_attribution_model.setter
    def external_attribution_model(self, value):
        self._external_attribution_model = value

    @property
    def external_attribution_credit(self):
        return self._external_attribution_credit;

    @external_attribution_credit.setter
    def external_attribution_credit(self, value):
        self._external_attribution_credit = value

    @property
    def offline_conversion(self):
        """ The offline conversion Data Object of the Campaign Management Service.

        """

        return self._offline_conversion

    @offline_conversion.setter
    def offline_conversion(self, value):
        self._offline_conversion = value

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.ConversionCurrencyCode,
            field_to_csv=lambda c: c.offline_conversion.ConversionCurrencyCode,
            csv_to_field=lambda c, v: setattr(
                c.offline_conversion,
                'ConversionCurrencyCode',
                v
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.ConversionName,
            field_to_csv=lambda c: c.offline_conversion.ConversionName,
            csv_to_field=lambda c, v: setattr(
                c.offline_conversion,
                'ConversionName',
                v
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.MicrosoftClickId,
            field_to_csv=lambda c: c.offline_conversion.MicrosoftClickId,
            csv_to_field=lambda c, v: setattr(
                c.offline_conversion,
                'MicrosoftClickId',
                v
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.ConversionValue,
            field_to_csv=lambda c: c.offline_conversion.ConversionValue,
            csv_to_field=lambda c, v: setattr(
                c.offline_conversion,
                'ConversionValue',
                float(v) if v else None
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.ConversionTime,
            field_to_csv=lambda c: bulk_datetime_str(c.offline_conversion.ConversionTime),
            csv_to_field=lambda c, v: setattr(
                c.offline_conversion,
                'ConversionTime',
                parse_datetime(v) if v else None
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.AdjustmentValue,
            field_to_csv=lambda c: c.adjustment_value,
            csv_to_field=lambda c, v: setattr(
                c,
                'adjustment_value',
                float(v) if v else None
            )
        ),

        _SimpleBulkMapping(
            header=_StringTable.AdjustmentType,
            field_to_csv=lambda c: c.adjustment_type,
            csv_to_field=lambda c, v: setattr(
                c,
                'adjustment_type',
                v
            )
        ),

        _SimpleBulkMapping(
            header=_StringTable.AdjustmentCurrencyCode,
            field_to_csv=lambda c: c.adjustment_currency_code,
            csv_to_field=lambda c, v: setattr(
                c,
                'adjustment_currency_code',
                v
            )
        ),

        _SimpleBulkMapping(
            header=_StringTable.ExternalAttributionModel,
            field_to_csv=lambda c: c.external_attribution_model,
            csv_to_field=lambda c, v: setattr(
                c,
                'external_attribution_model',
                v
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.ExternalAttributionCredit,
            field_to_csv=lambda c: c.external_attribution_credit,
            csv_to_field=lambda c, v: setattr(
                c,
                'external_attribution_credit',
                float(v) if v else None
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.AdjustmentTime,
            field_to_csv=lambda c: bulk_datetime_str(c.adjustment_time),
            csv_to_field=lambda c, v: setattr(
                c,
                'adjustment_time',
                parse_datetime(v) if v else None
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.HashedEmailAddress,
            field_to_csv=lambda c: c.offline_conversion.HashedEmailAddress,
            csv_to_field=lambda c, v: setattr(
                c.offline_conversion,
                'HashedEmailAddress',
                v
            )
        ),
        _SimpleBulkMapping(
            header=_StringTable.HashedPhoneNumber,
            field_to_csv=lambda c: c.offline_conversion.HashedPhoneNumber,
            csv_to_field=lambda c, v: setattr(
                c.offline_conversion,
                'HashedPhoneNumber',
                v
            )
        ),
    ]

    def process_mappings_to_row_values(self, row_values, exclude_readonly_data):
        self._validate_property_not_null(self._offline_conversion, 'offline_conversion')
        self.convert_to_values(row_values, BulkOfflineConversion._MAPPINGS)

    def process_mappings_from_row_values(self, row_values):
        self._offline_conversion = _CAMPAIGN_OBJECT_FACTORY_V13.create('OfflineConversion')
        row_values.convert_to_entity(self, BulkOfflineConversion._MAPPINGS)

    def read_additional_data(self, stream_reader):
        super(BulkOfflineConversion, self).read_additional_data(stream_reader)



# coding: utf-8

"""
    CampaignManagementService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union, Set
from typing_extensions import Self

class RecommendationCropSetting(BaseModel):
    """
    RecommendationCropSetting
    """ # noqa: E501
    source_x: Optional[StrictInt] = Field(default=None, alias="sourceX")
    source_y: Optional[StrictInt] = Field(default=None, alias="sourceY")
    source_width: Optional[StrictInt] = Field(default=None, alias="sourceWidth")
    source_height: Optional[StrictInt] = Field(default=None, alias="sourceHeight")
    aspect_ratio: Optional[StrictStr] = Field(default=None, alias="aspectRatio")
    __properties: ClassVar[List[str]] = ["sourceX", "sourceY", "sourceWidth", "sourceHeight", "aspectRatio"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )
	

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if source_x (nullable) is None
        # and model_fields_set contains the field
        if self.source_x is None and "source_x" in self.model_fields_set:
            _dict['sourceX'] = None

        # set to None if source_y (nullable) is None
        # and model_fields_set contains the field
        if self.source_y is None and "source_y" in self.model_fields_set:
            _dict['sourceY'] = None

        # set to None if source_width (nullable) is None
        # and model_fields_set contains the field
        if self.source_width is None and "source_width" in self.model_fields_set:
            _dict['sourceWidth'] = None

        # set to None if source_height (nullable) is None
        # and model_fields_set contains the field
        if self.source_height is None and "source_height" in self.model_fields_set:
            _dict['sourceHeight'] = None

        # set to None if aspect_ratio (nullable) is None
        # and model_fields_set contains the field
        if self.aspect_ratio is None and "aspect_ratio" in self.model_fields_set:
            _dict['aspectRatio'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of RecommendationCropSetting from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "sourceX": obj.get("SourceX") if obj.get("SourceX") is not None else None,
                        "sourceY": obj.get("SourceY") if obj.get("SourceY") is not None else None,
                        "sourceWidth": obj.get("SourceWidth") if obj.get("SourceWidth") is not None else None,
                        "sourceHeight": obj.get("SourceHeight") if obj.get("SourceHeight") is not None else None,
                        "aspectRatio": obj.get("AspectRatio") if obj.get("AspectRatio") is not None else None
        })
        return _obj



# coding: utf-8

"""
    CustomerManagementService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self

class Industry(str, Enum):
    NA = 'NA'

    AGENCYSALESHOUSE = 'AgencySalesHouse'

    AUTOMOTIVE = 'Automotive'

    CONSUMERPACKAGEDGOODS = 'ConsumerPackagedGoods'

    EDUCATION = 'Education'

    ENTERTAINMENT = 'Entertainment'

    FINANCIALSERVICES = 'FinancialServices'

    FOODSERVICES = 'FoodServices'

    GAMING = 'Gaming'

    GOVERNMENTNONPROFITPOLITICAL = 'GovernmentNonprofitPolitical'

    HEALTHCARE = 'Healthcare'

    INTERNAL = 'Internal'

    PUBLISHINGANDWEBMEDIA = 'PublishingAndWebMedia'

    REALESTATE = 'RealEstate'

    RETAIL = 'Retail'

    SERVICES = 'Services'

    TECHNOLOGY = 'Technology'

    TELECOMMUNICATIONS = 'Telecommunications'

    TRAVELHOSPITALITY = 'TravelHospitality'

    OTHER = 'Other'

    PHARMACEUTICALS = 'Pharmaceuticals'

	
    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of Industry from a JSON string"""
        return cls(json.loads(json_str))





from bingads.service_client import _CAMPAIGN_OBJECT_FACTORY_V13

from bingads.v13.internal.bulk.string_table import _StringTable
from bingads.v13.internal.bulk.entities.single_record_bulk_entity import _SingleRecordBulkEntity
from bingads.v13.internal.bulk.mappings import _SimpleBulkMapping, _ComplexBulkMapping
from bingads.v13.internal.extensions import *

class BulkAudienceGroupAssetGroupAssociation(_SingleRecordBulkEntity):
    """ Represents an audience group asset group association.

    This class exposes the property :attr:`audience_group_asset_group_association` that can be read and written as fields of the audience group asset group association record
    in a bulk file.

    For more information, see audience group asset group association at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self, status=None, audience_group_asset_group_association=None):
        super(BulkAudienceGroupAssetGroupAssociation, self).__init__()

        self._status = status
        self._audience_group_asset_group_association = audience_group_asset_group_association
        self._asset_group_name = None
        self._campaign_name = None

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self._status = status
    
    @property
    def asset_group_name(self):
        return self._asset_group_name

    @asset_group_name.setter
    def asset_group_name(self, asset_group_name):
        self._asset_group_name = asset_group_name
        
    @property
    def campaign_name(self):
        return self._campaign_name

    @campaign_name.setter
    def campaign_name(self, campaign_name):
        self._campaign_name = campaign_name

    @property
    def audience_group_asset_group_association(self):
        """ The AudienceGroupAssetGroupAssociation Data Object of the Campaign Management Service.

        A subset of AudienceGroupAssetGroupAssociation properties are available in the Ad Group record.
        For more information, see Ad Group at https://go.microsoft.com/fwlink/?linkid=846127.
        """
        return self._audience_group_asset_group_association

    @audience_group_asset_group_association.setter
    def audience_group_asset_group_association(self, audience_group_asset_group_association):
        self._audience_group_asset_group_association = audience_group_asset_group_association

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.Id,
            field_to_csv=lambda c: bulk_str(c.audience_group_asset_group_association.AudienceGroupId),
            csv_to_field=lambda c, v: setattr(c.audience_group_asset_group_association, 'AudienceGroupId', int(v) if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.Status,
            field_to_csv=lambda c: bulk_str(c.status),
            csv_to_field=lambda c, v: setattr(c, 'status', v if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.ParentId,
            field_to_csv=lambda c: bulk_str(c.audience_group_asset_group_association.AssetGroupId),
            csv_to_field=lambda c, v: setattr(c.audience_group_asset_group_association, 'AssetGroupId', int(v) if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.AssetGroup,
            field_to_csv=lambda c: c.asset_group_name,
            csv_to_field=lambda c, v: setattr(c, 'asset_group_name', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.Campaign,
            field_to_csv=lambda c: c.campaign_name,
            csv_to_field=lambda c, v: setattr(c, 'campaign_name', v)
        ),
    ]


    def process_mappings_from_row_values(self, row_values):
        self.audience_group_asset_group_association = _CAMPAIGN_OBJECT_FACTORY_V13.create('AudienceGroupAssetGroupAssociation')

        row_values.convert_to_entity(self, BulkAudienceGroupAssetGroupAssociation._MAPPINGS)

    def process_mappings_to_row_values(self, row_values, exclude_readonly_data):
        self._validate_property_not_null(self._audience_group_asset_group_association, 'AudienceGroupAssetGroupAssociation')
        self.convert_to_values(row_values, BulkAudienceGroupAssetGroupAssociation._MAPPINGS)

    def read_additional_data(self, stream_reader):
        super(BulkAudienceGroupAssetGroupAssociation, self).read_additional_data(stream_reader)



from .bulk_object import _BulkObject
from .string_table import _StringTable
from .mappings import _SimpleBulkMapping


class _FormatVersion(_BulkObject):
    def __init__(self):
        self._value = None

    @property
    def value(self):
        return self._value

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.Name,
            csv_to_field=lambda c, v: setattr(c, "_value", v)
        )
    ]

    def read_from_row_values(self, row_values):
        row_values.convert_to_entity(self, _FormatVersion._MAPPINGS)

    def write_to_row_values(self, row_values):
        super(_FormatVersion, self).write_to_row_values(row_values)

    def read_related_data_from_stream(self, stream_reader):
        super(_FormatVersion, self).read_related_data_from_stream(stream_reader)

    def write_to_stream(self, row_writer):
        super(_FormatVersion, self).write_to_stream(row_writer)

    @property
    def can_enclose_in_multiline_entity(self):
        return super(_FormatVersion, self).can_enclose_in_multiline_entity

    def enclose_in_multiline_entity(self):
        return super(_FormatVersion, self).enclose_in_multiline_entity()



from bingads.v13.internal.bulk.mappings import _SimpleBulkMapping
from bingads.v13.internal.bulk.string_table import _StringTable
from bingads.service_client import _CAMPAIGN_OBJECT_FACTORY_V13

from .common import _BulkAdExtensionBase
from .common import _BulkAdGroupAdExtensionAssociation
from .common import _BulkCampaignAdExtensionAssociation
from .common import _BulkAccountAdExtensionAssociation

from bingads.v13.internal.extensions import *


_PromotionAdExtension = type(_CAMPAIGN_OBJECT_FACTORY_V13.create('PromotionAdExtension'))


class BulkPromotionAdExtension(_BulkAdExtensionBase):
    """ Represents a promotion ad extension.

    This class exposes the :attr:`promotion_ad_extension` property that can be read and written
    as fields of the Promotion Ad Extension record in a bulk file.

    For more information, see Promotion Ad Extension at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self, account_id=None, ad_extension=None):
        if ad_extension and not isinstance(ad_extension, _PromotionAdExtension):
            raise ValueError('The type of ad_extension is: {0}, should be: {1}'.format(
                type(ad_extension),
                'PromotionAdExtension'
            ))
        super(BulkPromotionAdExtension, self).__init__(
            account_id=account_id,
            ad_extension=ad_extension
        )
        

    @property
    def promotion_ad_extension(self):
        """ The promotion ad extension.

        see Promotion Ad Extension at https://go.microsoft.com/fwlink/?linkid=846127.
        """

        return self._ad_extension

    @promotion_ad_extension.setter
    def promotion_ad_extension(self, value):
        self._ad_extension = value
        

    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.Language,
            field_to_csv=lambda c: bulk_str(c.promotion_ad_extension.Language),
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'Language', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.CurrencyCode,
            field_to_csv=lambda c: bulk_str(c.promotion_ad_extension.CurrencyCode),
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'CurrencyCode', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.PromotionCode,
            field_to_csv=lambda c: bulk_optional_str(c.promotion_ad_extension.PromotionCode, c.promotion_ad_extension.Id),
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'PromotionCode', v if v else '')
        ),
        _SimpleBulkMapping(
            header=_StringTable.PromotionTarget,
            field_to_csv=lambda c: bulk_str(c.promotion_ad_extension.PromotionItem),
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'PromotionItem', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.PromotionStart,
            field_to_csv=lambda c: field_to_csv_SchedulingDate(c.promotion_ad_extension.PromotionStartDate, c.promotion_ad_extension.Id) if c._ad_extension.Scheduling else None,
            csv_to_field=lambda c, v: csv_to_field_Date(c.promotion_ad_extension, 'PromotionStartDate', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.PromotionEnd,
            field_to_csv=lambda c: field_to_csv_SchedulingDate(c.promotion_ad_extension.PromotionEndDate, c.promotion_ad_extension.Id) if c._ad_extension.Scheduling else None,
            csv_to_field = lambda c, v: csv_to_field_Date(c.promotion_ad_extension, 'PromotionEndDate', v)
        ),        
        _SimpleBulkMapping(
            header=_StringTable.MoneyAmountOff,
            field_to_csv=lambda c: c.promotion_ad_extension.MoneyAmountOff,
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'MoneyAmountOff',float(v) if v else None)
        ),        
        _SimpleBulkMapping(
            header=_StringTable.OrdersOverAmount,
            field_to_csv=lambda c: c.promotion_ad_extension.OrdersOverAmount,
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'OrdersOverAmount',float(v) if v else None)
        ),        
        _SimpleBulkMapping(
            header=_StringTable.PercentOff,
            field_to_csv=lambda c: c.promotion_ad_extension.PercentOff,
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'PercentOff',float(v) if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.DiscountModifier,
            field_to_csv=lambda c: bulk_str(c.promotion_ad_extension.DiscountModifier),
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'DiscountModifier', v if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.Occasion,
            field_to_csv=lambda c: bulk_str(c.promotion_ad_extension.PromotionOccasion),
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'PromotionOccasion', v if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.FinalUrl,
            field_to_csv=lambda c: field_to_csv_Urls(c.promotion_ad_extension.FinalUrls, c.promotion_ad_extension.Id),
            csv_to_field=lambda c, v: csv_to_field_Urls(c.promotion_ad_extension.FinalUrls, v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.FinalMobileUrl,
            field_to_csv=lambda c: field_to_csv_Urls(c.promotion_ad_extension.FinalMobileUrls, c.promotion_ad_extension.Id),
            csv_to_field=lambda c, v: csv_to_field_Urls(c.promotion_ad_extension.FinalMobileUrls, v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.TrackingTemplate,
            field_to_csv=lambda c: bulk_optional_str(c.promotion_ad_extension.TrackingUrlTemplate, c.promotion_ad_extension.Id),
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'TrackingUrlTemplate', v if v else '')
        ),
        _SimpleBulkMapping(
            header=_StringTable.CustomParameter,
            field_to_csv=lambda c: field_to_csv_UrlCustomParameters(c.promotion_ad_extension),
            csv_to_field=lambda c, v: csv_to_field_UrlCustomParameters(c.promotion_ad_extension, v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.FinalUrlSuffix,
            field_to_csv=lambda c: bulk_optional_str(c.promotion_ad_extension.FinalUrlSuffix, c.promotion_ad_extension.Id),
            csv_to_field=lambda c, v: setattr(c.promotion_ad_extension, 'FinalUrlSuffix', v)
        )
    ]

    def process_mappings_from_row_values(self, row_values):
        self.promotion_ad_extension = _CAMPAIGN_OBJECT_FACTORY_V13.create('PromotionAdExtension')
        self.promotion_ad_extension.Type = 'PromotionAdExtension'
        super(BulkPromotionAdExtension, self).process_mappings_from_row_values(row_values)
        row_values.convert_to_entity(self, BulkPromotionAdExtension._MAPPINGS)

    def process_mappings_to_row_values(self, row_values, exclude_readonly_data):
        self._validate_property_not_null(self.promotion_ad_extension, 'promotion_ad_extension')
        super(BulkPromotionAdExtension, self).process_mappings_to_row_values(row_values, exclude_readonly_data)
        self.convert_to_values(row_values, BulkPromotionAdExtension._MAPPINGS)


class BulkAccountPromotionAdExtension(_BulkAccountAdExtensionAssociation):
    """ Represents an account level promotion ad extension.

    This class exposes properties that can be read and written
    as fields of the Account Promotion Ad Extension record in a bulk file.

    For more information, see Account Promotion Ad Extension at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    pass

class BulkCampaignPromotionAdExtension(_BulkCampaignAdExtensionAssociation):
    """ Represents a campaign level promotion ad extension.

    This class exposes properties that can be read and written
    as fields of the Campaign Promotion Ad Extension record in a bulk file.

    For more information, see Campaign Promotion Ad Extension at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    pass

class BulkAdGroupPromotionAdExtension(_BulkAdGroupAdExtensionAssociation):
    """ Represents an ad group level Promotion ad extension.

    This class exposes properties that can be read and written
    as fields of the Ad Group Promotion Ad Extension record in a bulk file.

    For more information, see Ad Group Promotion Ad Extension at https://go.microsoft.com/fwlink/?linkid=846127.

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    pass



# coding: utf-8

"""
    CampaignManagementService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union, Set
from openapi_client.models.campaign.ad_editorial_status import AdEditorialStatus
from openapi_client.models.campaign.ad_status import AdStatus
from openapi_client.models.campaign.app_url import AppUrl
from openapi_client.models.campaign.custom_parameters import CustomParameters
from openapi_client.models.campaign.key_value_pair_ofstring_andstring import KeyValuePairOfstringAndstring
from typing_extensions import Self

class DynamicSearchAd(BaseModel):
    """
    DynamicSearchAd
    """ # noqa: E501
    text: Optional[StrictStr] = Field(default=None, alias="Text")
    text_part2: Optional[StrictStr] = Field(default=None, alias="TextPart2")
    path1: Optional[StrictStr] = Field(default=None, alias="Path1")
    path2: Optional[StrictStr] = Field(default=None, alias="Path2")
    id: Optional[StrictStr] = Field(default=None, alias="Id")
    type: Optional[StrictStr] = Field(default='DynamicSearch', alias="Type")
    status: Optional[AdStatus] = Field(default=None, alias="Status")
    editorial_status: Optional[AdEditorialStatus] = Field(default=None, alias="EditorialStatus")
    device_preference: Optional[StrictStr] = Field(default=None, alias="DevicePreference")
    ad_format_preference: Optional[StrictStr] = Field(default=None, alias="AdFormatPreference")
    tracking_url_template: Optional[StrictStr] = Field(default=None, alias="TrackingUrlTemplate")
    final_url_suffix: Optional[StrictStr] = Field(default=None, alias="FinalUrlSuffix")
    url_custom_parameters: Optional[CustomParameters] = Field(default=None, alias="UrlCustomParameters")
    final_urls: Optional[List[StrictStr]] = Field(default=None, alias="FinalUrls")
    final_mobile_urls: Optional[List[StrictStr]] = Field(default=None, alias="FinalMobileUrls")
    final_app_urls: Optional[List[Optional[AppUrl]]] = Field(default=None, alias="FinalAppUrls")
    forward_compatibility_map: Optional[List[Optional[KeyValuePairOfstringAndstring]]] = Field(default=None, alias="ForwardCompatibilityMap")
    __properties: ClassVar[List[str]] = ["Text", "TextPart2", "Path1", "Path2", "Id", "Type", "Status", "EditorialStatus", "DevicePreference", "AdFormatPreference", "TrackingUrlTemplate", "FinalUrlSuffix", "UrlCustomParameters", "FinalUrls", "FinalMobileUrls", "FinalAppUrls", "ForwardCompatibilityMap"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )
	
    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())
	
    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of DynamicSearchAd from a JSON string"""
        return cls.from_dict(json.loads(json_str))
	

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of url_custom_parameters
        if self.url_custom_parameters:
            _dict['UrlCustomParameters'] = self.url_custom_parameters.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in final_app_urls (list)
        _items = []
        if self.final_app_urls:
            for _item_final_app_urls in self.final_app_urls:
                if _item_final_app_urls:
                    _items.append(_item_final_app_urls.to_dict())
            _dict['FinalAppUrls'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in forward_compatibility_map (list)
        _items = []
        if self.forward_compatibility_map:
            for _item_forward_compatibility_map in self.forward_compatibility_map:
                if _item_forward_compatibility_map:
                    _items.append(_item_forward_compatibility_map.to_dict())
            _dict['ForwardCompatibilityMap'] = _items
        # set to None if text (nullable) is None
        # and model_fields_set contains the field
        if self.text is None and "text" in self.model_fields_set:
            _dict['Text'] = None

        # set to None if text_part2 (nullable) is None
        # and model_fields_set contains the field
        if self.text_part2 is None and "text_part2" in self.model_fields_set:
            _dict['TextPart2'] = None

        # set to None if path1 (nullable) is None
        # and model_fields_set contains the field
        if self.path1 is None and "path1" in self.model_fields_set:
            _dict['Path1'] = None

        # set to None if path2 (nullable) is None
        # and model_fields_set contains the field
        if self.path2 is None and "path2" in self.model_fields_set:
            _dict['Path2'] = None

        # set to None if id (nullable) is None
        # and model_fields_set contains the field
        if self.id is None and "id" in self.model_fields_set:
            _dict['Id'] = None

        # set to None if type (nullable) is None
        # and model_fields_set contains the field
        if self.type is None and "type" in self.model_fields_set:
            _dict['Type'] = None

        # set to None if status (nullable) is None
        # and model_fields_set contains the field
        if self.status is None and "status" in self.model_fields_set:
            _dict['Status'] = None

        # set to None if editorial_status (nullable) is None
        # and model_fields_set contains the field
        if self.editorial_status is None and "editorial_status" in self.model_fields_set:
            _dict['EditorialStatus'] = None

        # set to None if device_preference (nullable) is None
        # and model_fields_set contains the field
        if self.device_preference is None and "device_preference" in self.model_fields_set:
            _dict['DevicePreference'] = None

        # set to None if ad_format_preference (nullable) is None
        # and model_fields_set contains the field
        if self.ad_format_preference is None and "ad_format_preference" in self.model_fields_set:
            _dict['AdFormatPreference'] = None

        # set to None if tracking_url_template (nullable) is None
        # and model_fields_set contains the field
        if self.tracking_url_template is None and "tracking_url_template" in self.model_fields_set:
            _dict['TrackingUrlTemplate'] = None

        # set to None if final_url_suffix (nullable) is None
        # and model_fields_set contains the field
        if self.final_url_suffix is None and "final_url_suffix" in self.model_fields_set:
            _dict['FinalUrlSuffix'] = None

        # set to None if url_custom_parameters (nullable) is None
        # and model_fields_set contains the field
        if self.url_custom_parameters is None and "url_custom_parameters" in self.model_fields_set:
            _dict['UrlCustomParameters'] = None

        # set to None if final_urls (nullable) is None
        # and model_fields_set contains the field
        if self.final_urls is None and "final_urls" in self.model_fields_set:
            _dict['FinalUrls'] = None

        # set to None if final_mobile_urls (nullable) is None
        # and model_fields_set contains the field
        if self.final_mobile_urls is None and "final_mobile_urls" in self.model_fields_set:
            _dict['FinalMobileUrls'] = None

        # set to None if final_app_urls (nullable) is None
        # and model_fields_set contains the field
        if self.final_app_urls is None and "final_app_urls" in self.model_fields_set:
            _dict['FinalAppUrls'] = None

        # set to None if forward_compatibility_map (nullable) is None
        # and model_fields_set contains the field
        if self.forward_compatibility_map is None and "forward_compatibility_map" in self.model_fields_set:
            _dict['ForwardCompatibilityMap'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of DynamicSearchAd from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "Text": obj.get("Text") if obj.get("Text") is not None else None,
                        "TextPart2": obj.get("TextPart2") if obj.get("TextPart2") is not None else None,
                        "Path1": obj.get("Path1") if obj.get("Path1") is not None else None,
                        "Path2": obj.get("Path2") if obj.get("Path2") is not None else None,
                        "Id": obj.get("Id") if obj.get("Id") is not None else None,
                        "Type": obj.get("Type") if obj.get("Type") is not None else 'DynamicSearch',
                        "Status": obj.get("Status") if obj.get("Status") is not None else None,
                        "EditorialStatus": obj.get("EditorialStatus") if obj.get("EditorialStatus") is not None else None,
                        "DevicePreference": obj.get("DevicePreference") if obj.get("DevicePreference") is not None else None,
                        "AdFormatPreference": obj.get("AdFormatPreference") if obj.get("AdFormatPreference") is not None else None,
                        "TrackingUrlTemplate": obj.get("TrackingUrlTemplate") if obj.get("TrackingUrlTemplate") is not None else None,
                        "FinalUrlSuffix": obj.get("FinalUrlSuffix") if obj.get("FinalUrlSuffix") is not None else None,
                        "UrlCustomParameters": CustomParameters.from_dict(obj["UrlCustomParameters"]) if obj.get("UrlCustomParameters") is not None else None,
                        "FinalUrls": obj.get("FinalUrls"),
                        "FinalMobileUrls": obj.get("FinalMobileUrls"),
                        "FinalAppUrls": [AppUrl.from_dict(_item) for _item in obj["FinalAppUrls"]] if obj.get("FinalAppUrls") is not None else None,
                        "ForwardCompatibilityMap": [KeyValuePairOfstringAndstring.from_dict(_item) for _item in obj["ForwardCompatibilityMap"]] if obj.get("ForwardCompatibilityMap") is not None else None
        })
        return _obj



# coding: utf-8

"""
    CustomerManagementService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union, Set
from openapi_client.models.customer.key_value_entity_ofstring_andstring import KeyValueEntityOfstringAndstring
from typing_extensions import Self

class Notification(BaseModel):
    """
    Notification
    """ # noqa: E501
    type_id: Optional[StrictInt] = Field(default=None, alias="TypeId")
    id: Optional[StrictStr] = Field(default=None, alias="Id")
    severity: Optional[StrictInt] = Field(default=None, alias="Severity")
    title: Optional[StrictStr] = Field(default=None, alias="Title")
    message: Optional[StrictStr] = Field(default=None, alias="Message")
    additional_info: Optional[List[Optional[KeyValueEntityOfstringAndstring]]] = Field(default=None, alias="AdditionalInfo")
    __properties: ClassVar[List[str]] = ["TypeId", "Id", "Severity", "Title", "Message", "AdditionalInfo"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )
	

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in additional_info (list)
        _items = []
        if self.additional_info:
            for _item_additional_info in self.additional_info:
                if _item_additional_info:
                    _items.append(_item_additional_info.to_dict())
            _dict['AdditionalInfo'] = _items
        # set to None if type_id (nullable) is None
        # and model_fields_set contains the field
        if self.type_id is None and "type_id" in self.model_fields_set:
            _dict['TypeId'] = None

        # set to None if id (nullable) is None
        # and model_fields_set contains the field
        if self.id is None and "id" in self.model_fields_set:
            _dict['Id'] = None

        # set to None if severity (nullable) is None
        # and model_fields_set contains the field
        if self.severity is None and "severity" in self.model_fields_set:
            _dict['Severity'] = None

        # set to None if title (nullable) is None
        # and model_fields_set contains the field
        if self.title is None and "title" in self.model_fields_set:
            _dict['Title'] = None

        # set to None if message (nullable) is None
        # and model_fields_set contains the field
        if self.message is None and "message" in self.model_fields_set:
            _dict['Message'] = None

        # set to None if additional_info (nullable) is None
        # and model_fields_set contains the field
        if self.additional_info is None and "additional_info" in self.model_fields_set:
            _dict['AdditionalInfo'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of Notification from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "TypeId": obj.get("TypeId") if obj.get("TypeId") is not None else None,
                        "Id": obj.get("Id") if obj.get("Id") is not None else None,
                        "Severity": obj.get("Severity") if obj.get("Severity") is not None else None,
                        "Title": obj.get("Title") if obj.get("Title") is not None else None,
                        "Message": obj.get("Message") if obj.get("Message") is not None else None,
                        "AdditionalInfo": [KeyValueEntityOfstringAndstring.from_dict(_item) for _item in obj["AdditionalInfo"]] if obj.get("AdditionalInfo") is not None else None
        })
        return _obj



class BulkOperationStatus(object):
    """ Contains tracking details about the results and status of the corresponding :class:`.BulkDownloadOperation` or :class:`.BulkUploadOperation`. """

    def __init__(self,
                 status=None,
                 percent_complete=None,
                 result_file_url=None,
                 errors=None):
        """ Initialize a new instance of this class.

        :param status: (optional) The download or upload status.
        :type status: str
        :param percent_complete: (optional) Percent complete progress information for the bulk operation.
        :type percent_complete: int
        :param result_file_url: (optional) The download or upload result file Url.
        :type result_file_url: str
        :param errors: (optional) The list of errors associated with the operation.
        :type errors: list[OperationError]
        """

        self._status = status
        self._percent_complete = percent_complete
        self._result_file_url = result_file_url
        self._errors = errors

    @property
    def status(self):
        """ The download or upload status.

        :rtype: str
        """

        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    @property
    def percent_complete(self):
        """ Percent complete progress information for the bulk operation.

        :rtype: int
        """

        return self._percent_complete

    @percent_complete.setter
    def percent_complete(self, value):
        self._percent_complete = value

    @property
    def result_file_url(self):
        """ The download or upload result file Url.

        :rtype: str
        """

        return self._result_file_url

    @result_file_url.setter
    def result_file_url(self, value):
        self._result_file_url = value

    @property
    def errors(self):
        """ The list of errors associated with the operation.

        :rtype: list[OperationError]
        """

        return self._errors

    @errors.setter
    def errors(self, value):
        self._errors = value



from enum import Enum

class ResultFileType(Enum):
    """ Defines the possible types of result files.  """

    full_download = 1
    """ The result file represents the full sync of entities that were specified in the download request. """

    partial_download = 2
    """ The result file represents the partial sync of entities that were specified in the download request. """

    upload = 3
    """ The result file represents the entities specified in the upload request, or the corresponding errors, or both entities and errors. """



# coding: utf-8

# flake8: noqa
"""
    BulkService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


# import models into model package
from openapi_client.models.bulk.ad_api_error import AdApiError
from openapi_client.models.bulk.ad_api_fault_detail import AdApiFaultDetail
from openapi_client.models.bulk.ad_center_authentication_fault import AdCenterAuthenticationFault
from openapi_client.models.bulk.ad_center_authorization_fault import AdCenterAuthorizationFault
from openapi_client.models.bulk.ad_center_fault import AdCenterFault
from openapi_client.models.bulk.ad_center_internal_error_fault import AdCenterInternalErrorFault
from openapi_client.models.bulk.ad_center_system_fault import AdCenterSystemFault
from openapi_client.models.bulk.api_fault_detail import ApiFaultDetail
from openapi_client.models.bulk.application_fault import ApplicationFault
from openapi_client.models.bulk.authentication_fault_code import AuthenticationFaultCode
from openapi_client.models.bulk.batch_error import BatchError
from openapi_client.models.bulk.batch_error_base import BatchErrorBase
from openapi_client.models.bulk.campaign_scope import CampaignScope
from openapi_client.models.bulk.compression_type import CompressionType
from openapi_client.models.bulk.data_scope import DataScope
from openapi_client.models.bulk.download_campaigns_by_account_ids_request import DownloadCampaignsByAccountIdsRequest
from openapi_client.models.bulk.download_campaigns_by_account_ids_response import DownloadCampaignsByAccountIdsResponse
from openapi_client.models.bulk.download_campaigns_by_campaign_ids_request import DownloadCampaignsByCampaignIdsRequest
from openapi_client.models.bulk.download_campaigns_by_campaign_ids_response import DownloadCampaignsByCampaignIdsResponse
from openapi_client.models.bulk.download_entity import DownloadEntity
from openapi_client.models.bulk.download_file_type import DownloadFileType
from openapi_client.models.bulk.editorial_api_fault_detail import EditorialApiFaultDetail
from openapi_client.models.bulk.editorial_error import EditorialError
from openapi_client.models.bulk.fault_source import FaultSource
from openapi_client.models.bulk.get_bulk_download_status_request import GetBulkDownloadStatusRequest
from openapi_client.models.bulk.get_bulk_download_status_response import GetBulkDownloadStatusResponse
from openapi_client.models.bulk.get_bulk_upload_status_request import GetBulkUploadStatusRequest
from openapi_client.models.bulk.get_bulk_upload_status_response import GetBulkUploadStatusResponse
from openapi_client.models.bulk.get_bulk_upload_url_request import GetBulkUploadUrlRequest
from openapi_client.models.bulk.get_bulk_upload_url_response import GetBulkUploadUrlResponse
from openapi_client.models.bulk.key_value_pair_ofstring_andstring import KeyValuePairOfstringAndstring
from openapi_client.models.bulk.operation_error import OperationError
from openapi_client.models.bulk.response_mode import ResponseMode
from openapi_client.models.bulk.upload_entity_records_request import UploadEntityRecordsRequest
from openapi_client.models.bulk.upload_entity_records_response import UploadEntityRecordsResponse



from bingads.service_client import _CAMPAIGN_OBJECT_FACTORY_V13
from bingads.v13.internal.bulk.string_table import _StringTable
from bingads.v13.internal.bulk.entities.single_record_bulk_entity import _SingleRecordBulkEntity
from bingads.v13.internal.bulk.mappings import _SimpleBulkMapping
from bingads.v13.internal.extensions import *


class BulkCampaignConversionGoal(_SingleRecordBulkEntity):
    """ Represents a CampaignConversionGoal.

    Properties of this class and of classes that it is derived from, correspond to fields of the CampaignConversionGoal record in a bulk file.
    For more information, see CampaignConversionGoal at https://go.microsoft.com/fwlink/?linkid=846127

    *See also:*

    * :class:`.BulkServiceManager`
    * :class:`.BulkOperation`
    * :class:`.BulkFileReader`
    * :class:`.BulkFileWriter`
    """

    def __init__(self, campaign_conversion_goal = None, sub_type = None, action_type = None):
        super(BulkCampaignConversionGoal, self).__init__()
        self._campaign_conversion_goal = campaign_conversion_goal
        self._sub_type = sub_type
        self._action_type = action_type

    @property
    def campaign_conversion_goal(self):
        """ define a campaign conversion goal association

        :rtype: CampaignConversionGoal
        """
        return self._campaign_conversion_goal

    @campaign_conversion_goal.setter
    def campaign_conversion_goal(self, value):
        self._campaign_conversion_goal = value

    @property
    def sub_type(self):
        """ Corresponds to the 'Sub Type' field in the bulk file.

        :rtype: str
        """
        return self._sub_type

    @sub_type.setter
    def sub_type(self, value):
        self._sub_type = value

    @property
    def action_type(self):
        """ Corresponds to the 'Action Type' field in the bulk file.

        :rtype: str
        """
        return self._action_type

    @action_type.setter
    def action_type(self, value):
        self._action_type = value


    _MAPPINGS = [
        _SimpleBulkMapping(
            header=_StringTable.ParentId,
            field_to_csv=lambda c: bulk_str(c.campaign_conversion_goal.CampaignId),
            csv_to_field=lambda c, v: setattr(c.campaign_conversion_goal, 'CampaignId', int(v) if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.GoalId,
            field_to_csv=lambda c: bulk_str(c.campaign_conversion_goal.GoalId),
            csv_to_field=lambda c, v: setattr(c.campaign_conversion_goal, 'GoalId', int(v) if v else None)
        ),
        _SimpleBulkMapping(
            header=_StringTable.ActionType,
            field_to_csv=lambda c: c.action_type,
            csv_to_field=lambda c, v: setattr(c, 'action_type', v)
        ),
        _SimpleBulkMapping(
            header=_StringTable.SubType,
            field_to_csv=lambda c: c.sub_type,
            csv_to_field=lambda c, v: setattr(c, 'sub_type', v)
        ),
    ]

    def process_mappings_from_row_values(self, row_values):
        self._campaign_conversion_goal = _CAMPAIGN_OBJECT_FACTORY_V13.create('CampaignConversionGoal')
        row_values.convert_to_entity(self, BulkCampaignConversionGoal._MAPPINGS)

    def process_mappings_to_row_values(self, row_values, exclude_readonly_data):
        self.convert_to_values(row_values, BulkCampaignConversionGoal._MAPPINGS)

    def read_additional_data(self, stream_reader):
        super(BulkCampaignConversionGoal, self).read_additional_data(stream_reader)



import csv
from .bulk_object_factory import _BulkObjectFactory
from .row_values import _RowValues
from .csv_reader import _CsvReader

class _BulkObjectReader(object):
    """ Provides a method to read one row from bulk file and return the corresponding :class:`._BulkObject` """

    def __init__(self, _csv_reader):
        self._csv_reader = _csv_reader
        self._csv_reader.__enter__()
        headers = self._read_headers()
        self._column_mapping = dict(zip(headers, range(0, len(headers))))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._csv_reader.__exit__(exc_type, exc_value, traceback)

    def __iter__(self):
        return self

    def __next__(self):
        return self.read_next_bulk_object()

    def next(self):
        return self.__next__()

    def read_next_bulk_object(self):
        """ Reads the next csv row values, creates a new instance of the object and populates it with the row values.

        :return: next bulk object
        :rtype: _BulkObject
        """
        try:
            row_values = self._read_next_row_values()
        except StopIteration:
            return None
        bulk_object = _BulkObjectFactory.create_bulk_object(row_values)
        bulk_object.read_from_row_values(row_values)
        return bulk_object

    def close(self):
        self.__exit__(None, None, None)

    def _read_next_row_values(self):
        values = next(self._csv_reader)
        return _RowValues(columns=values, mappings=self._column_mapping)

    def _read_headers(self):
        # Need to strip BOM marker by hand, take care
        def remove_bom(unicode_str):
            unicode_bom = u'\N{ZERO WIDTH NO-BREAK SPACE}'
            if unicode_str and unicode_str[0] == unicode_bom:
                unicode_str = unicode_str[1:]
            return unicode_str

        headers = next(self._csv_reader)
        return [remove_bom(header) for header in headers]

class _BulkFileObjectReader(_BulkObjectReader):
    """ Provides a method to read one row from bulk file and return the corresponding :class:`._BulkObject` """

    def __init__(self, file_path, delimiter, encoding='utf-8-sig'):
        super(_BulkFileObjectReader, self).__init__(_CsvReader(filename=file_path, delimiter=delimiter, encoding=encoding))
        self._file_path = file_path
        self._delimiter = delimiter
        self._encoding = encoding


    @property
    def file_path(self):
        return self._file_path

    @property
    def delimiter(self):
        return self._delimiter

    @property
    def encoding(self):
        return self._encoding

class _CsvRowsReader:
    def __init__(self, csv_rows):
        self._csv_reader = csv.reader(csv_rows, dialect=csv.excel)
        
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass
    
    def __iter__(self):
        return self

    def __next__(self):
        return self.next()
    
    def next(self):
        return next(self._csv_reader)

class _BulkRowsObjectReader(_BulkObjectReader):
    """ Provides a method to read one row from bulk file and return the corresponding :class:`._BulkObject` """

    def __init__(self, csv_rows):
        super(_BulkRowsObjectReader, self).__init__(_CsvRowsReader(csv_rows))


# coding: utf-8

"""
    CampaignManagementService

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union, Set
from openapi_client.models.campaign.ad_group_privacy_status import AdGroupPrivacyStatus
from openapi_client.models.campaign.ad_group_status import AdGroupStatus
from openapi_client.models.campaign.ad_rotation import AdRotation
from openapi_client.models.campaign.bid import Bid
from openapi_client.models.campaign.bidding_scheme import BiddingScheme
from openapi_client.models.campaign.custom_parameters import CustomParameters
from openapi_client.models.campaign.frequency_cap_settings import FrequencyCapSettings
from openapi_client.models.campaign.key_value_pair_ofstring_andstring import KeyValuePairOfstringAndstring
from openapi_client.models.campaign.model_date import ModelDate
from openapi_client.models.campaign.network import Network
from openapi_client.models.campaign.rate_bid import RateBid
from openapi_client.models.campaign.setting import Setting
from typing_extensions import Self

class AdGroup(BaseModel):
    """
    AdGroup
    """ # noqa: E501
    id: Optional[StrictStr] = Field(default=None, alias="Id")
    name: Optional[StrictStr] = Field(default=None, alias="Name")
    start_date: Optional[ModelDate] = Field(default=None, alias="StartDate")
    end_date: Optional[ModelDate] = Field(default=None, alias="EndDate")
    language: Optional[StrictStr] = Field(default=None, alias="Language")
    status: Optional[AdGroupStatus] = Field(default=None, alias="Status")
    privacy_status: Optional[AdGroupPrivacyStatus] = Field(default=None, alias="PrivacyStatus")
    cpc_bid: Optional[Bid] = Field(default=None, alias="CpcBid")
    network: Optional[Network] = Field(default=None, alias="Network")
    ad_rotation: Optional[AdRotation] = Field(default=None, alias="AdRotation")
    audience_ads_bid_adjustment: Optional[StrictInt] = Field(default=None, alias="AudienceAdsBidAdjustment")
    tracking_url_template: Optional[StrictStr] = Field(default=None, alias="TrackingUrlTemplate")
    final_url_suffix: Optional[StrictStr] = Field(default=None, alias="FinalUrlSuffix")
    url_custom_parameters: Optional[CustomParameters] = Field(default=None, alias="UrlCustomParameters")
    bidding_scheme: Optional[BiddingScheme] = Field(default=None, alias="BiddingScheme")
    settings: Optional[List[Optional[Setting]]] = Field(default=None, alias="Settings")
    ad_schedule_use_searcher_time_zone: Optional[StrictBool] = Field(default=None, alias="AdScheduleUseSearcherTimeZone")
    ad_group_type: Optional[StrictStr] = Field(default=None, alias="AdGroupType")
    cpv_bid: Optional[Bid] = Field(default=None, alias="CpvBid")
    cpm_bid: Optional[Bid] = Field(default=None, alias="CpmBid")
    mcpa_bid: Optional[Bid] = Field(default=None, alias="McpaBid")
    multimedia_ads_bid_adjustment: Optional[StrictInt] = Field(default=None, alias="MultimediaAdsBidAdjustment")
    commission_rate: Optional[RateBid] = Field(default=None, alias="CommissionRate")
    percent_cpc_bid: Optional[RateBid] = Field(default=None, alias="PercentCpcBid")
    use_optimized_targeting: Optional[StrictBool] = Field(default=None, alias="UseOptimizedTargeting")
    use_predictive_targeting: Optional[StrictBool] = Field(default=None, alias="UsePredictiveTargeting")
    frequency_cap_settings: Optional[List[Optional[FrequencyCapSettings]]] = Field(default=None, alias="FrequencyCapSettings")
    forward_compatibility_map: Optional[List[Optional[KeyValuePairOfstringAndstring]]] = Field(default=None, alias="ForwardCompatibilityMap")
    __properties: ClassVar[List[str]] = ["Id", "Name", "StartDate", "EndDate", "Language", "Status", "PrivacyStatus", "CpcBid", "Network", "AdRotation", "AudienceAdsBidAdjustment", "TrackingUrlTemplate", "FinalUrlSuffix", "UrlCustomParameters", "BiddingScheme", "Settings", "AdScheduleUseSearcherTimeZone", "AdGroupType", "CpvBid", "CpmBid", "McpaBid", "MultimediaAdsBidAdjustment", "CommissionRate", "PercentCpcBid", "UseOptimizedTargeting", "UsePredictiveTargeting", "FrequencyCapSettings", "ForwardCompatibilityMap"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )
	

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of start_date
        if self.start_date:
            _dict['StartDate'] = self.start_date.to_dict()
        # override the default output from pydantic by calling `to_dict()` of end_date
        if self.end_date:
            _dict['EndDate'] = self.end_date.to_dict()
        # override the default output from pydantic by calling `to_dict()` of cpc_bid
        if self.cpc_bid:
            _dict['CpcBid'] = self.cpc_bid.to_dict()
        # override the default output from pydantic by calling `to_dict()` of ad_rotation
        if self.ad_rotation:
            _dict['AdRotation'] = self.ad_rotation.to_dict()
        # override the default output from pydantic by calling `to_dict()` of url_custom_parameters
        if self.url_custom_parameters:
            _dict['UrlCustomParameters'] = self.url_custom_parameters.to_dict()
        # override the default output from pydantic by calling `to_dict()` of bidding_scheme
        if self.bidding_scheme:
            _dict['BiddingScheme'] = self.bidding_scheme.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in settings (list)
        _items = []
        if self.settings:
            for _item_settings in self.settings:
                if _item_settings:
                    _items.append(_item_settings.to_dict())
            _dict['Settings'] = _items
        # override the default output from pydantic by calling `to_dict()` of cpv_bid
        if self.cpv_bid:
            _dict['CpvBid'] = self.cpv_bid.to_dict()
        # override the default output from pydantic by calling `to_dict()` of cpm_bid
        if self.cpm_bid:
            _dict['CpmBid'] = self.cpm_bid.to_dict()
        # override the default output from pydantic by calling `to_dict()` of mcpa_bid
        if self.mcpa_bid:
            _dict['McpaBid'] = self.mcpa_bid.to_dict()
        # override the default output from pydantic by calling `to_dict()` of commission_rate
        if self.commission_rate:
            _dict['CommissionRate'] = self.commission_rate.to_dict()
        # override the default output from pydantic by calling `to_dict()` of percent_cpc_bid
        if self.percent_cpc_bid:
            _dict['PercentCpcBid'] = self.percent_cpc_bid.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in frequency_cap_settings (list)
        _items = []
        if self.frequency_cap_settings:
            for _item_frequency_cap_settings in self.frequency_cap_settings:
                if _item_frequency_cap_settings:
                    _items.append(_item_frequency_cap_settings.to_dict())
            _dict['FrequencyCapSettings'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in forward_compatibility_map (list)
        _items = []
        if self.forward_compatibility_map:
            for _item_forward_compatibility_map in self.forward_compatibility_map:
                if _item_forward_compatibility_map:
                    _items.append(_item_forward_compatibility_map.to_dict())
            _dict['ForwardCompatibilityMap'] = _items
        # set to None if id (nullable) is None
        # and model_fields_set contains the field
        if self.id is None and "id" in self.model_fields_set:
            _dict['Id'] = None

        # set to None if name (nullable) is None
        # and model_fields_set contains the field
        if self.name is None and "name" in self.model_fields_set:
            _dict['Name'] = None

        # set to None if start_date (nullable) is None
        # and model_fields_set contains the field
        if self.start_date is None and "start_date" in self.model_fields_set:
            _dict['StartDate'] = None

        # set to None if end_date (nullable) is None
        # and model_fields_set contains the field
        if self.end_date is None and "end_date" in self.model_fields_set:
            _dict['EndDate'] = None

        # set to None if language (nullable) is None
        # and model_fields_set contains the field
        if self.language is None and "language" in self.model_fields_set:
            _dict['Language'] = None

        # set to None if status (nullable) is None
        # and model_fields_set contains the field
        if self.status is None and "status" in self.model_fields_set:
            _dict['Status'] = None

        # set to None if privacy_status (nullable) is None
        # and model_fields_set contains the field
        if self.privacy_status is None and "privacy_status" in self.model_fields_set:
            _dict['PrivacyStatus'] = None

        # set to None if cpc_bid (nullable) is None
        # and model_fields_set contains the field
        if self.cpc_bid is None and "cpc_bid" in self.model_fields_set:
            _dict['CpcBid'] = None

        # set to None if network (nullable) is None
        # and model_fields_set contains the field
        if self.network is None and "network" in self.model_fields_set:
            _dict['Network'] = None

        # set to None if ad_rotation (nullable) is None
        # and model_fields_set contains the field
        if self.ad_rotation is None and "ad_rotation" in self.model_fields_set:
            _dict['AdRotation'] = None

        # set to None if audience_ads_bid_adjustment (nullable) is None
        # and model_fields_set contains the field
        if self.audience_ads_bid_adjustment is None and "audience_ads_bid_adjustment" in self.model_fields_set:
            _dict['AudienceAdsBidAdjustment'] = None

        # set to None if tracking_url_template (nullable) is None
        # and model_fields_set contains the field
        if self.tracking_url_template is None and "tracking_url_template" in self.model_fields_set:
            _dict['TrackingUrlTemplate'] = None

        # set to None if final_url_suffix (nullable) is None
        # and model_fields_set contains the field
        if self.final_url_suffix is None and "final_url_suffix" in self.model_fields_set:
            _dict['FinalUrlSuffix'] = None

        # set to None if url_custom_parameters (nullable) is None
        # and model_fields_set contains the field
        if self.url_custom_parameters is None and "url_custom_parameters" in self.model_fields_set:
            _dict['UrlCustomParameters'] = None

        # set to None if bidding_scheme (nullable) is None
        # and model_fields_set contains the field
        if self.bidding_scheme is None and "bidding_scheme" in self.model_fields_set:
            _dict['BiddingScheme'] = None

        # set to None if settings (nullable) is None
        # and model_fields_set contains the field
        if self.settings is None and "settings" in self.model_fields_set:
            _dict['Settings'] = None

        # set to None if ad_schedule_use_searcher_time_zone (nullable) is None
        # and model_fields_set contains the field
        if self.ad_schedule_use_searcher_time_zone is None and "ad_schedule_use_searcher_time_zone" in self.model_fields_set:
            _dict['AdScheduleUseSearcherTimeZone'] = None

        # set to None if ad_group_type (nullable) is None
        # and model_fields_set contains the field
        if self.ad_group_type is None and "ad_group_type" in self.model_fields_set:
            _dict['AdGroupType'] = None

        # set to None if cpv_bid (nullable) is None
        # and model_fields_set contains the field
        if self.cpv_bid is None and "cpv_bid" in self.model_fields_set:
            _dict['CpvBid'] = None

        # set to None if cpm_bid (nullable) is None
        # and model_fields_set contains the field
        if self.cpm_bid is None and "cpm_bid" in self.model_fields_set:
            _dict['CpmBid'] = None

        # set to None if mcpa_bid (nullable) is None
        # and model_fields_set contains the field
        if self.mcpa_bid is None and "mcpa_bid" in self.model_fields_set:
            _dict['McpaBid'] = None

        # set to None if multimedia_ads_bid_adjustment (nullable) is None
        # and model_fields_set contains the field
        if self.multimedia_ads_bid_adjustment is None and "multimedia_ads_bid_adjustment" in self.model_fields_set:
            _dict['MultimediaAdsBidAdjustment'] = None

        # set to None if commission_rate (nullable) is None
        # and model_fields_set contains the field
        if self.commission_rate is None and "commission_rate" in self.model_fields_set:
            _dict['CommissionRate'] = None

        # set to None if percent_cpc_bid (nullable) is None
        # and model_fields_set contains the field
        if self.percent_cpc_bid is None and "percent_cpc_bid" in self.model_fields_set:
            _dict['PercentCpcBid'] = None

        # set to None if use_optimized_targeting (nullable) is None
        # and model_fields_set contains the field
        if self.use_optimized_targeting is None and "use_optimized_targeting" in self.model_fields_set:
            _dict['UseOptimizedTargeting'] = None

        # set to None if use_predictive_targeting (nullable) is None
        # and model_fields_set contains the field
        if self.use_predictive_targeting is None and "use_predictive_targeting" in self.model_fields_set:
            _dict['UsePredictiveTargeting'] = None

        # set to None if frequency_cap_settings (nullable) is None
        # and model_fields_set contains the field
        if self.frequency_cap_settings is None and "frequency_cap_settings" in self.model_fields_set:
            _dict['FrequencyCapSettings'] = None

        # set to None if forward_compatibility_map (nullable) is None
        # and model_fields_set contains the field
        if self.forward_compatibility_map is None and "forward_compatibility_map" in self.model_fields_set:
            _dict['ForwardCompatibilityMap'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of AdGroup from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "Id": obj.get("Id") if obj.get("Id") is not None else None,
                        "Name": obj.get("Name") if obj.get("Name") is not None else None,
                        "StartDate": ModelDate.from_dict(obj["StartDate"]) if obj.get("StartDate") is not None else None,
                        "EndDate": ModelDate.from_dict(obj["EndDate"]) if obj.get("EndDate") is not None else None,
                        "Language": obj.get("Language") if obj.get("Language") is not None else None,
                        "Status": obj.get("Status") if obj.get("Status") is not None else None,
                        "PrivacyStatus": obj.get("PrivacyStatus") if obj.get("PrivacyStatus") is not None else None,
                        "CpcBid": Bid.from_dict(obj["CpcBid"]) if obj.get("CpcBid") is not None else None,
                        "Network": obj.get("Network") if obj.get("Network") is not None else None,
                        "AdRotation": AdRotation.from_dict(obj["AdRotation"]) if obj.get("AdRotation") is not None else None,
                        "AudienceAdsBidAdjustment": obj.get("AudienceAdsBidAdjustment") if obj.get("AudienceAdsBidAdjustment") is not None else None,
                        "TrackingUrlTemplate": obj.get("TrackingUrlTemplate") if obj.get("TrackingUrlTemplate") is not None else None,
                        "FinalUrlSuffix": obj.get("FinalUrlSuffix") if obj.get("FinalUrlSuffix") is not None else None,
                        "UrlCustomParameters": CustomParameters.from_dict(obj["UrlCustomParameters"]) if obj.get("UrlCustomParameters") is not None else None,
                        "BiddingScheme": BiddingScheme.from_dict(obj["BiddingScheme"]) if obj.get("BiddingScheme") is not None else None,
                        "Settings": [Setting.from_dict(_item) for _item in obj["Settings"]] if obj.get("Settings") is not None else None,
                        "AdScheduleUseSearcherTimeZone": obj.get("AdScheduleUseSearcherTimeZone") if obj.get("AdScheduleUseSearcherTimeZone") is not None else None,
                        "AdGroupType": obj.get("AdGroupType") if obj.get("AdGroupType") is not None else None,
                        "CpvBid": Bid.from_dict(obj["CpvBid"]) if obj.get("CpvBid") is not None else None,
                        "CpmBid": Bid.from_dict(obj["CpmBid"]) if obj.get("CpmBid") is not None else None,
                        "McpaBid": Bid.from_dict(obj["McpaBid"]) if obj.get("McpaBid") is not None else None,
                        "MultimediaAdsBidAdjustment": obj.get("MultimediaAdsBidAdjustment") if obj.get("MultimediaAdsBidAdjustment") is not None else None,
                        "CommissionRate": RateBid.from_dict(obj["CommissionRate"]) if obj.get("CommissionRate") is not None else None,
                        "PercentCpcBid": RateBid.from_dict(obj["PercentCpcBid"]) if obj.get("PercentCpcBid") is not None else None,
                        "UseOptimizedTargeting": obj.get("UseOptimizedTargeting") if obj.get("UseOptimizedTargeting") is not None else None,
                        "UsePredictiveTargeting": obj.get("UsePredictiveTargeting") if obj.get("UsePredictiveTargeting") is not None else None,
                        "FrequencyCapSettings": [FrequencyCapSettings.from_dict(_item) for _item in obj["FrequencyCapSettings"]] if obj.get("FrequencyCapSettings") is not None else None,
                        "ForwardCompatibilityMap": [KeyValuePairOfstringAndstring.from_dict(_item) for _item in obj["ForwardCompatibilityMap"]] if obj.get("ForwardCompatibilityMap") is not None else None
        })
        return _obj



import csv
import codecs


class _CsvWriter:
    def __init__(self, filename, delimiter):
        self._filename = filename
        self._delimiter = delimiter
        self._encoding = 'utf-8-sig'

        if delimiter == ',':
            self._dialect = csv.excel
        elif delimiter == '\t':
            self._dialect = csv.excel_tab
        else:
            raise ValueError('Do not support delimiter: {0}', delimiter)

        self._csv_file = codecs.open(filename, mode='w', encoding=self._encoding)
        self._csv_writer = csv.writer(self._csv_file, dialect=self._dialect)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._csv_file.flush()
        self._csv_file.close()

    def close(self):
        self.__exit__(None, None, None)

    def writerow(self, row):
        self._csv_writer.writerow(row)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)

    @property
    def filename(self):
        return self._filename

    @property
    def delimiter(self):
        return self._delimiter
